from typing import Any, Literal

from pydantic import AnyHttpUrl, BaseModel, Field, FilePath


class PathInput(BaseModel):
    """Input referencing a file path on the server filesystem."""

    type: Literal["path"] = Field(description="Input type discriminator")
    path: FilePath = Field(description="Absolute or relative path to file on server filesystem")


class UrlInput(BaseModel):
    """Input referencing a remote URL."""

    type: Literal["url"] = Field(description="Input type discriminator")
    url: AnyHttpUrl = Field(description="URL to fetch the resource from")


class ContentInput(BaseModel):
    """Input with inline content."""

    type: Literal["content"] = Field(description="Input type discriminator")
    content: str = Field(description="Inline content as string")


class FileContentInput(BaseModel):
    """Input with file-backed content and original filename."""

    type: Literal["file_content"] = Field(description="Input type discriminator")
    filename: str = Field(description="Original filename provided by client")
    content: str = Field(description="Inline file content as string")


BaseInput = PathInput | ContentInput
ConfigInput = BaseInput
SchemaInput = BaseInput | FileContentInput | UrlInput


class BaseExportRequest(BaseModel):
    """Base request model for all export endpoints with common parameters."""

    schemas: list[SchemaInput] = Field(
        description="Array of schema inputs (paths or URLs)",
        json_schema_extra={"x-property-format": "graphql", "x-cli-flag": "--schema"},
    )
    selection_query: ConfigInput | None = Field(
        default=None,
        description="Optional GraphQL query for filtering (path or inline content)",
        json_schema_extra={"x-property-format": "graphql", "x-cli-flag": "--selection-query"},
    )
    naming_config: ConfigInput | None = Field(
        default=None,
        description="Optional YAML naming configuration (path or inline content)",
        json_schema_extra={
            "x-property-format": "yaml",
            "x-cli-flag": "--naming-config",
            "x-docs-url": "https://covesa.github.io/s2dm/docs/tools/command-line-interface-cli/#naming-configuration",
        },
    )
    root_type: str | None = Field(
        default=None,
        description="Optional root type name for filtering the schema",
        json_schema_extra={"x-cli-flag": "--root-type"},
    )
    expanded_instances: bool = Field(
        default=False,
        description="Whether to expand instance tags into nested structures",
        json_schema_extra={"x-cli-flag": "--expanded-instances"},
    )


class QueryBasedExportRequest(BaseExportRequest):
    """Export request that requires a selection query."""

    selection_query: ConfigInput = Field(
        description="GraphQL query for filtering (path or inline content)",
        json_schema_extra={"x-property-format": "graphql", "x-cli-flag": "--selection-query"},
    )


class ResponseMetadata(BaseModel):
    """Metadata for API responses."""

    result_format: str = Field(description="Format of the result content")
    processing_time_ms: int | None = Field(default=None, description="Processing time in milliseconds")


class ApiResponse(BaseModel):
    """Standard response model for API endpoints."""

    result: list[str] = Field(description="Array of result content strings (schemas, exports, etc.)")
    metadata: ResponseMetadata | None = Field(
        default=None, description="Response metadata (only present if there is a result)"
    )


class ErrorResponse(BaseModel):
    """Error response model."""

    error: str = Field(description="Error type or category")
    message: str = Field(description="Detailed error message")
    details: dict[str, Any] | None = Field(default=None, description="Additional error context")



from typing import cast

from graphql import (
    GraphQLEnumType,
    GraphQLField,
    GraphQLList,
    GraphQLNamedType,
    GraphQLNonNull,
    GraphQLNullableType,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLType,
    GraphQLUnionType,
    get_named_type,
    is_enum_type,
    is_list_type,
    is_non_null_type,
    is_object_type,
    is_scalar_type,
    is_union_type,
)

from s2dm import log
from s2dm.exporters.utils.annotated_schema import AnnotatedSchema

from .common import get_avro_scalar_type


class AvroProtocolTransformer:
    """
    Transformer class to convert GraphQL schema to Avro protocol format.
    """

    def __init__(
        self,
        annotated_schema: AnnotatedSchema,
        namespace: str,
        protocol_name: str,
        referenced_types: set[GraphQLNamedType],
        strict: bool = False,
    ):
        self.annotated_schema = annotated_schema
        self.graphql_schema = annotated_schema.schema
        self.namespace = namespace
        self.protocol_name = protocol_name
        self.referenced_types = referenced_types
        self.strict = strict
        self.processed_types: set[str] = set()

    def transform(self) -> str:
        """
        Transform GraphQL types to Avro IDL protocol.

        Returns:
            str: Avro IDL protocol definition
        """
        lines = [f'@namespace("{self.namespace}")', f"protocol {self.protocol_name} {{"]

        for graphql_type in self.referenced_types:
            type_name = graphql_type.name

            if type_name in self.processed_types:
                continue

            self.processed_types.add(type_name)

            if is_object_type(graphql_type):
                lines.append(self._transform_record(cast(GraphQLObjectType, graphql_type)))
            elif is_enum_type(graphql_type) and self.strict:
                lines.append(self._transform_enum(cast(GraphQLEnumType, graphql_type)))
            elif is_union_type(graphql_type):
                lines.append(self._transform_union(cast(GraphQLUnionType, graphql_type)))

        lines.append("}")

        return "\n".join(lines)

    def _transform_record(self, obj_type: GraphQLObjectType) -> str:
        """Transform GraphQL object type to Avro record."""
        lines = [f"  record {obj_type.name} {{"]

        for field_name, field in obj_type.fields.items():
            field_type_str = self._get_field_type_string(field)
            comment = self._get_field_comment(field)
            if comment:
                lines.append(f"    {field_type_str} {field_name}; {comment}")
            else:
                lines.append(f"    {field_type_str} {field_name};")

        lines.append("  }")

        return "\n".join(lines)

    def _transform_enum(self, enum_type: GraphQLEnumType) -> str:
        """Transform GraphQL enum to Avro enum."""
        values = ", ".join(enum_type.values.keys())
        return f"  enum {enum_type.name} {{ {values} }}"

    def _transform_union(self, union_type: GraphQLUnionType) -> str:
        """Transform GraphQL union to Avro record with union field."""
        lines = [f"  record {union_type.name} {{"]

        union_types = [t.name for t in union_type.types]
        union_str = "union { " + ", ".join(union_types) + " }"
        lines.append(f"    {union_str} value;")

        lines.append("  }")

        return "\n".join(lines)

    def _get_field_type_string(self, field: GraphQLField) -> str:
        """Get Avro IDL type string for a GraphQL field."""
        field_type: GraphQLType = field.type
        is_required = is_non_null_type(field_type)

        if is_required:
            field_type = cast(GraphQLNonNull[GraphQLNullableType], field_type).of_type

        avro_type = self._get_avro_type(field_type, field)

        if self.strict and is_required:
            return avro_type
        return f"{avro_type}?"

    def _get_avro_type(self, field_type: GraphQLType, field: GraphQLField | None = None) -> str:
        """Get Avro IDL type for a GraphQL field type."""
        if is_list_type(field_type):
            list_type = cast(GraphQLList[GraphQLType], field_type)
            element_type: GraphQLType = list_type.of_type

            if is_non_null_type(element_type):
                element_type = cast(GraphQLNonNull[GraphQLNullableType], element_type).of_type

            element_type_str = self._get_avro_type(element_type)
            return f"array<{element_type_str}>"

        if is_scalar_type(field_type):
            scalar_type = cast(GraphQLScalarType, field_type)
            return get_avro_scalar_type(scalar_type, field)

        if is_enum_type(field_type):
            enum_type = cast(GraphQLEnumType, field_type)
            return enum_type.name if self.strict else "string"

        if is_object_type(field_type) or is_union_type(field_type):
            named_type = cast(GraphQLNamedType, field_type)
            return named_type.name

        log.warning(f"Unsupported GraphQL type '{field_type}' for Avro IDL protocol conversion, using string")
        return "string"

    def _get_field_comment(self, field: GraphQLField) -> str:
        """Get comment for enum fields in non-strict mode."""
        if self.strict:
            return ""

        base_type = get_named_type(field.type)
        if is_enum_type(base_type):
            return f"/* enum {base_type.name} */"

        return ""



"""Query validate route - validate GraphQL query against schema."""

from fastapi import APIRouter
from graphql import parse, print_schema, validate

from s2dm.api.config import COMMON_RESPONSES
from s2dm.api.errors import ResponseError, format_error_list
from s2dm.api.models.base import ApiResponse
from s2dm.api.models.query_validate import ValidateQueryRequest
from s2dm.api.services.response_service import execute_and_respond
from s2dm.api.services.schema_service import path_for_content, process_schema_input
from s2dm.exporters.utils.schema_loader import load_schema

router = APIRouter(responses=COMMON_RESPONSES)


@router.post("/validate", response_model=ApiResponse)
def validate_query(request: ValidateQueryRequest) -> ApiResponse:
    """Validate a GraphQL query against a schema. Returns the schema if validation succeeds."""

    def process_request() -> list[str]:
        schema_paths = [process_schema_input(schema_input) for schema_input in request.schemas]

        schema = load_schema(schema_paths)

        query_path = path_for_content(request.selection_query, "selection_query", ".graphql")
        query_text = query_path.read_text(encoding="utf-8")

        query_document = parse(query_text)

        validation_errors = validate(schema, query_document)

        if validation_errors:
            error_messages = [error.message for error in validation_errors]
            raise ResponseError(format_error_list("Query validation failed", error_messages))

        schema_content = print_schema(schema)
        return [schema_content]

    return execute_and_respond(executor=process_request, result_format="graphql")



"""Models for query validate endpoint."""

from pydantic import BaseModel, Field

from s2dm.api.models.base import ConfigInput, SchemaInput


class ValidateQueryRequest(BaseModel):
    """Request model for validating a GraphQL query against a schema."""

    schemas: list[SchemaInput] = Field(..., description="GraphQL schema files to validate against")
    selection_query: ConfigInput = Field(..., description="GraphQL query to validate")



"""End-to-end integration tests for JSON Schema export."""

import json
import tempfile
from pathlib import Path

import pytest

from s2dm.exporters.jsonschema import translate_to_jsonschema
from s2dm.exporters.jsonschema.jsonschema import transform
from s2dm.exporters.utils.schema_loader import load_and_process_schema


class TestJsonSchemaE2E:
    @pytest.fixture
    def test_schema_path(self, spec_directory: Path) -> list[Path]:
        """Path to the test GraphQL schema."""
        return [spec_directory, Path(__file__).parent / "test_expanded_instances" / "test_schema.graphql"]

    def test_root_node_filters_types(self, tmp_path: Path) -> None:
        """Test that root node filtering only includes reachable types."""
        schema_str = """
            type Query { vehicle: Vehicle, engine: Engine }
            type Vehicle { id: ID!, engine: Engine }
            type Engine { id: ID!, displacement: Float! }
            type UnrelatedType { id: ID!, data: String }
        """
        schema_file = tmp_path / "test_root_node_filters_types.graphql"
        schema_file.write_text(schema_str)

        root_type = "Vehicle"
        annotated_schema, _, _ = load_and_process_schema(
            schema_paths=[schema_file],
            naming_config_path=None,
            selection_query_path=None,
            root_type=root_type,
            expanded_instances=False,
        )

        result = transform(annotated_schema.schema, root_type=root_type, strict=False)
        schema = json.loads(result)

        assert "Vehicle" in schema["$defs"]
        assert "Engine" in schema["$defs"]
        assert schema["title"] == "Vehicle"
        assert schema["$ref"] == "#/$defs/Vehicle"

        assert "UnrelatedType" not in schema["$defs"]

    def test_instance_tag_expansion(self, test_schema_path: list[Path]) -> None:
        """Test expanded instances for seats with 3-level nesting are inlined in parent property."""
        root_type = "Cabin"
        annotated_schema, _, _ = load_and_process_schema(
            schema_paths=test_schema_path,
            naming_config_path=None,
            selection_query_path=None,
            root_type=root_type,
            expanded_instances=True,
        )
        result = translate_to_jsonschema(annotated_schema, root_type=root_type)
        schema = json.loads(result)

        # Intermediate types must NOT appear as named $defs
        assert "Seat_Row" not in schema["$defs"]
        assert "Seat_Position" not in schema["$defs"]

        # The expanded structure is inlined directly in the parent property
        cabin_def = schema["$defs"]["Cabin"]
        seat_property = cabin_def["properties"]["Seat"]
        assert seat_property["type"] == "object"
        assert seat_property["additionalProperties"] is False
        assert set(seat_property["properties"].keys()) == {"ROW1", "ROW2", "ROW3"}

        row1 = seat_property["properties"]["ROW1"]
        assert row1["type"] == "object"
        assert set(row1["properties"].keys()) == {"LEFT", "CENTER", "RIGHT"}
        assert row1["properties"]["LEFT"]["$ref"] == "#/$defs/Seat"
        assert row1["properties"]["CENTER"]["$ref"] == "#/$defs/Seat"
        assert row1["properties"]["RIGHT"]["$ref"] == "#/$defs/Seat"

        seat_def = schema["$defs"]["Seat"]
        assert "isOccupied" in seat_def["properties"]

    def test_default_behavior_creates_arrays(self, test_schema_path: list[Path]) -> None:
        """Test that the default behavior creates arrays for instance tagged objects."""
        root_type = "Cabin"
        annotated_schema, _, _ = load_and_process_schema(
            schema_paths=test_schema_path,
            naming_config_path=None,
            selection_query_path=None,
            root_type=root_type,
            expanded_instances=False,
        )
        result = translate_to_jsonschema(annotated_schema, root_type=root_type)
        schema = json.loads(result)

        doors_def = schema["$defs"]["Cabin"]["properties"]["doors"]
        assert doors_def["type"] == "array"
        assert "items" in doors_def

        seats_def = schema["$defs"]["Cabin"]["properties"]["seats"]
        assert seats_def["type"] == "array"
        assert "items" in seats_def

    def test_non_instance_tagged_objects_remain_arrays(
        self, spec_directory: Path, test_schema_path: list[Path]
    ) -> None:
        """Test that objects without instance tags remain as arrays even with expanded_instances=True."""
        # Create a schema with both instance-tagged and regular arrays
        extended_schema = """
        enum RowEnum {
          ROW1
          ROW2
        }

        enum SideEnum {
          DRIVERSIDE
          PASSENGERSIDE
        }

        type DoorPosition @instanceTag {
          Row: RowEnum!
          Side: SideEnum!
        }

        type Door {
          isLocked: Boolean
          instanceTag: DoorPosition
        }

        type RegularItem {
          name: String
          value: Int
        }

        type TestObject {
          doorsWithInstanceTag: [Door] @noDuplicates
          regularItems: [RegularItem] @noDuplicates
        }
        """

        with tempfile.NamedTemporaryFile(mode="w", suffix=".graphql", delete=False) as f:
            f.write(extended_schema)
            temp_path = Path(f.name)

        try:
            root_type = "TestObject"
            annotated_schema, _, _ = load_and_process_schema(
                schema_paths=[spec_directory, temp_path],
                naming_config_path=None,
                selection_query_path=None,
                root_type=root_type,
                expanded_instances=True,
            )
            result = translate_to_jsonschema(annotated_schema, root_type=root_type)
            schema = json.loads(result)

            # The expanded Door structure is inlined (no named intermediate $defs)
            doors_def = schema["$defs"]["TestObject"]["properties"]["Door"]
            assert doors_def["type"] == "object"
            assert doors_def["additionalProperties"] is False
            assert set(doors_def["properties"].keys()) == {"ROW1", "ROW2"}

            row1 = doors_def["properties"]["ROW1"]
            assert row1["type"] == "object"
            assert set(row1["properties"].keys()) == {"DRIVERSIDE", "PASSENGERSIDE"}
            assert row1["properties"]["DRIVERSIDE"]["$ref"] == "#/$defs/Door"

            assert "Door_Row" not in schema["$defs"]

            # Regular arrays remain unaffected
            items_def = schema["$defs"]["TestObject"]["properties"]["regularItems"]
            assert items_def["type"] == "array"
            assert "items" in items_def

        finally:
            temp_path.unlink()

    def test_expanded_instances_with_strict_mode(self, test_schema_path: list[Path]) -> None:
        """Test that expanded instances are inlined correctly with strict mode."""
        root_type = "Cabin"
        annotated_schema, _, _ = load_and_process_schema(
            schema_paths=test_schema_path,
            naming_config_path=None,
            selection_query_path=None,
            root_type=root_type,
            expanded_instances=True,
        )
        result = translate_to_jsonschema(annotated_schema, root_type=root_type, strict=True)
        schema = json.loads(result)

        # Intermediate types must NOT appear as named $defs
        assert "Door_Row" not in schema["$defs"]
        assert "Door_Side" not in schema["$defs"]

        # The expansion is inlined in the parent property
        cabin_def = schema["$defs"]["Cabin"]
        door_property = cabin_def["properties"]["Door"]
        assert door_property["type"] == "object"
        assert door_property["additionalProperties"] is False
        assert set(door_property["properties"].keys()) == {"ROW1", "ROW2"}

        row1 = door_property["properties"]["ROW1"]
        assert row1["type"] == "object"
        assert set(row1["properties"].keys()) == {"DRIVERSIDE", "PASSENGERSIDE"}

        # In strict mode, nullable leaf references use oneOf with null
        driverside_field = row1["properties"]["DRIVERSIDE"]
        assert "oneOf" in driverside_field
        refs = [item.get("$ref") for item in driverside_field["oneOf"] if "$ref" in item]
        assert "#/$defs/Door" in refs

    def test_singular_naming_for_expanded_instances(self, test_schema_path: list[Path]) -> None:
        """Test that expanded instances use singular type names instead of field names."""
        root_type = "Cabin"
        annotated_schema_normal, _, _ = load_and_process_schema(
            schema_paths=test_schema_path,
            naming_config_path=None,
            selection_query_path=None,
            root_type=root_type,
            expanded_instances=False,
        )
        result_normal = translate_to_jsonschema(annotated_schema_normal, root_type=root_type)

        annotated_schema_expanded, _, _ = load_and_process_schema(
            schema_paths=test_schema_path,
            naming_config_path=None,
            selection_query_path=None,
            root_type=root_type,
            expanded_instances=True,
        )
        result_expanded = translate_to_jsonschema(annotated_schema_expanded, root_type=root_type)

        schema_normal = json.loads(result_normal)
        schema_expanded = json.loads(result_expanded)

        assert "doors" in schema_normal["$defs"]["Cabin"]["properties"]
        assert "seats" in schema_normal["$defs"]["Cabin"]["properties"]

        assert "Door" in schema_expanded["$defs"]["Cabin"]["properties"]
        assert "Seat" in schema_expanded["$defs"]["Cabin"]["properties"]

        assert "doors" not in schema_expanded["$defs"]["Cabin"]["properties"]
        assert "seats" not in schema_expanded["$defs"]["Cabin"]["properties"]

    def test_nested_instances_use_inline_expansion(self, spec_directory: Path) -> None:
        """Test that nested expanded instances are inlined at each level without named intermediate $defs."""
        nested_schema_path = Path(__file__).parent / "test_expanded_instances" / "test_nested_schema.graphql"

        root_type = "Chassis"
        annotated_schema, _, _ = load_and_process_schema(
            schema_paths=[spec_directory, nested_schema_path],
            naming_config_path=None,
            selection_query_path=None,
            root_type=root_type,
            expanded_instances=True,
        )
        result = translate_to_jsonschema(annotated_schema, root_type=root_type)
        schema = json.loads(result)

        # Intermediate types must NOT appear as named $defs
        assert "Axle_Row" not in schema["$defs"]
        assert "Wheel_Position" not in schema["$defs"]

        # Chassis.Axle is inlined
        chassis_def = schema["$defs"]["Chassis"]
        axle_prop = chassis_def["properties"]["Axle"]
        assert axle_prop["type"] == "object"
        assert axle_prop["additionalProperties"] is False
        assert set(axle_prop["properties"].keys()) == {"ROW1", "ROW2"}
        assert axle_prop["properties"]["ROW1"]["$ref"] == "#/$defs/Axle"
        assert axle_prop["properties"]["ROW2"]["$ref"] == "#/$defs/Axle"

        # Axle.Wheel is also inlined
        axle_def = schema["$defs"]["Axle"]
        wheel_prop = axle_def["properties"]["Wheel"]
        assert wheel_prop["type"] == "object"
        assert wheel_prop["additionalProperties"] is False
        assert set(wheel_prop["properties"].keys()) == {"LEFT", "RIGHT"}
        assert wheel_prop["properties"]["LEFT"]["$ref"] == "#/$defs/Wheel"
        assert wheel_prop["properties"]["RIGHT"]["$ref"] == "#/$defs/Wheel"

        # Field names are singular
        assert "Wheel" in axle_def["properties"]
        assert "Wheels" not in axle_def["properties"]
        assert "Axle" in chassis_def["properties"]
        assert "Axles" not in chassis_def["properties"]

        wheel_def = schema["$defs"]["Wheel"]
        assert wheel_def["type"] == "object"
        assert "Tire" in wheel_def["properties"]
        assert wheel_def["properties"]["Tire"]["$ref"] == "#/$defs/Tire"



from itertools import product
from typing import Any, cast

from graphql import (
    GraphQLEnumType,
    GraphQLField,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    get_named_type,
    is_list_type,
    is_non_null_type,
)

from s2dm import log
from s2dm.exporters.utils.annotated_schema import FieldMetadata, TypeMetadata
from s2dm.exporters.utils.directive import has_given_directive
from s2dm.exporters.utils.extraction import get_all_object_types, get_all_objects_with_directive
from s2dm.exporters.utils.naming import apply_naming_to_instance_values, convert_name
from s2dm.exporters.utils.naming_config import ContextType, ElementType, NamingConventionConfig, get_case_for_element


def is_instance_tag_field(field_name: str) -> bool:
    return field_name == "instanceTag"


def get_all_expanded_instance_tags(
    schema: GraphQLSchema,
    naming_config: NamingConventionConfig | None = None,
) -> dict[GraphQLObjectType, list[str]]:
    all_expanded_instance_tags: dict[GraphQLObjectType, list[str]] = {}
    for object in get_all_objects_with_directive(get_all_object_types(schema), "instanceTag"):
        all_expanded_instance_tags[object] = expand_instance_tag(object, naming_config)

    log.debug(f"All expanded tags in the spec: {all_expanded_instance_tags}")

    return all_expanded_instance_tags


def expand_instance_tag(object: GraphQLObjectType, naming_config: NamingConventionConfig | None = None) -> list[str]:
    log.debug(f"Expanding instanceTag for object: {object.name}")
    expanded_tags = []
    if not has_given_directive(object, "instanceTag"):
        raise ValueError(f"Object '{object.name}' does not have an instance tag directive.")
    else:
        tags_per_enum_field = []
        for field_name, field in object.fields.items():
            field_type = field.type
            if isinstance(field.type, GraphQLNonNull):
                field_type = get_named_type(field.type)
            if not isinstance(field_type, GraphQLEnumType):
                # TODO: Move this check to a validation function for the @instanceTag directive
                raise TypeError(f"Field '{field_name}' in object '{object.name}' is not an enum.")

            enum_values = apply_naming_to_instance_values(list(field_type.values.keys()), naming_config)
            tags_per_enum_field.append(enum_values)
        log.debug(f"Tags per field: {tags_per_enum_field}")

        # Combine tags from different enum fields
        for combination in product(*tags_per_enum_field):
            expanded_tags.append(".".join(combination))  # <-- Character separator can be changed HERE

        log.debug(f"Expanded tags: {expanded_tags}")

        return expanded_tags


def is_valid_instance_tag_field(field: GraphQLField, schema: GraphQLSchema) -> bool:
    """
    Check if the output type of the given field is a valid instanceTag.
    A valid instanceTag is an object type with the @instanceTag directive.

    Args:
        field (GraphQLField): The field to check.
        schema (GraphQLSchema): The GraphQL schema to validate against.

    Returns:
        bool: True if the field's output type is a valid instanceTag, False otherwise.
    """
    output_type = schema.get_type(get_named_type(field.type).name)
    return isinstance(output_type, GraphQLObjectType) and has_given_directive(output_type, "instanceTag")


def has_valid_instance_tag_field(object_type: GraphQLObjectType, schema: GraphQLSchema) -> bool:
    """
    Check if a given object type has a field named 'instanceTag' and if the field's output type
    is a valid instanceTag.

    Args:
        object_type (GraphQLObjectType): The object type to check.
        schema (GraphQLSchema): The GraphQL schema to validate against.

    Returns:
        bool: True if the object type has a valid instanceTag field, False otherwise.
    """
    if "instanceTag" in object_type.fields:
        log.debug(f"instanceTag? {True}")
        field = object_type.fields["instanceTag"]
        return is_valid_instance_tag_field(field, schema)
    else:
        log.debug(f"instanceTag? {False}")
        return False


def get_instance_tag_object(object_type: GraphQLObjectType, schema: GraphQLSchema) -> GraphQLObjectType | None:
    """
    Get the valid instance tag object type used in a valid instance tag field.

    Args:
        object_type (GraphQLObjectType): The object type to check.
        schema (GraphQLSchema): The GraphQL schema to validate against.

    Returns:
        GraphQLObjectType | None: The valid instance tag object type if found, None otherwise.
    """

    instance_tag_field_name = "instanceTag"
    if instance_tag_field_name in object_type.fields:
        field = object_type.fields[instance_tag_field_name]
        instance_tag_type = schema.get_type(get_named_type(field.type).name)
        if isinstance(
            instance_tag_type, GraphQLObjectType
        ):  # and has_given_directive(instance_tag_type, instance_tag_field_name):
            return instance_tag_type
    return None


def get_instance_tag_dict(
    instance_tag_object: GraphQLObjectType,
) -> dict[str, list[str]]:
    """
    Given a valid instance tag object type, return the list of all enum values by level.

    Args:
        instance_tag_object (GraphQLObjectType): The instance tag object type to process.

    Returns:
        dict[str, list[str]]: A dictionary where keys are field names and values are lists of enum values.
    """
    instance_tag_dict = {}

    for field_name, field in instance_tag_object.fields.items():
        field_type = field.type
        if is_non_null_type(field_type):
            field_type = cast(GraphQLNonNull[Any], field_type).of_type

        if isinstance(field_type, GraphQLEnumType):
            instance_tag_dict[field_name] = list(field_type.values.keys())
        else:
            raise TypeError(f"Field '{field_name}' in object '{instance_tag_object.name}' is not an enum.")

    return instance_tag_dict


def is_expandable_field(field: GraphQLField, schema: GraphQLSchema) -> bool:
    """
    Check if a field is expandable (has a base type that has instanceTag).

    Args:
        field: The GraphQL field to check
        schema: The GraphQL schema to validate against

    Returns:
        True if the field is expandable, False otherwise
    """
    base_type = get_named_type(field.type)
    if isinstance(base_type, GraphQLObjectType):
        return has_valid_instance_tag_field(base_type, schema)

    return False


def _collect_expandable_fields(
    schema: GraphQLSchema,
) -> list[tuple[GraphQLObjectType, str]]:
    """
    Collect all fields in the schema that need instance expansion.

    Args:
        schema: The GraphQL schema to scan

    Returns:
        List of tuples: (parent_type, field_name)
    """
    expandable_fields = []
    all_object_types = get_all_object_types(schema)

    for object_type in all_object_types:
        for field_name, field in object_type.fields.items():
            if is_expandable_field(field, schema):
                expandable_fields.append((object_type, field_name))

    return expandable_fields


def _create_intermediate_types(
    base_type: GraphQLObjectType,
    instance_tag_dict: dict[str, list[str]],
    list_item_nullable: bool,
) -> list[GraphQLObjectType]:
    """
    Create intermediate GraphQL types for instance expansion.

    Args:
        base_type: The base type (e.g., Door)
        instance_tag_dict: Dict of enum field names to their values
        list_item_nullable: Whether the original list items were nullable

    Returns:
        List of intermediate types, with first intermediate type at the end
    """
    enum_fields = list(instance_tag_dict.keys())
    intermediate_types: list[GraphQLObjectType] = []

    for i in range(len(enum_fields) - 1, -1, -1):
        enum_field_name = enum_fields[i]
        enum_values = instance_tag_dict[enum_field_name]

        intermediate_type_name = f"{base_type.name}_{enum_field_name.capitalize()}"

        is_leaf_level = i == len(enum_fields) - 1
        target_type = base_type if is_leaf_level else intermediate_types[-1]

        intermediate_fields = {}
        for enum_value in enum_values:
            field_type: GraphQLObjectType | GraphQLNonNull[GraphQLObjectType] = (
                target_type if is_leaf_level and list_item_nullable else GraphQLNonNull(target_type)
            )
            intermediate_fields[enum_value] = GraphQLField(field_type)

        intermediate_type = GraphQLObjectType(
            name=intermediate_type_name,
            fields=intermediate_fields,
        )

        intermediate_types.append(intermediate_type)
        log.debug(f"Created intermediate type '{intermediate_type_name}' with fields: {list(enum_values)}")

    return intermediate_types


def expand_instances_in_schema(
    schema: GraphQLSchema,
    naming_config: NamingConventionConfig | None = None,
) -> tuple[GraphQLSchema, dict[str, "TypeMetadata"], dict[tuple[str, str], "FieldMetadata"]]:
    """
    Expand instance-tagged fields in a GraphQL schema into nested object structures.

    For fields with types that contain instanceTag fields, this function creates intermediate
    GraphQL types representing each level of the instance tag hierarchy and modifies the
    parent field to use the singular type name.

    Args:
        schema: The GraphQL schema to modify
        naming_config: Optional naming configuration for transforming type and field names

    Returns:
        Tuple of (modified schema, type metadata dict, field metadata dict)
    """
    log.info("Starting instance expansion in schema")

    expandable_fields = _collect_expandable_fields(schema)
    log.info(f"Found {len(expandable_fields)} expandable fields")

    base_types_to_clean: set[GraphQLObjectType] = set()
    instance_tag_types_to_remove: set[str] = set()
    new_types: dict[str, GraphQLObjectType] = {}
    type_metadata: dict[str, TypeMetadata] = {}
    field_metadata: dict[tuple[str, str], FieldMetadata] = {}

    # TODO: Optimization - Cache expanded types to avoid creating duplicate intermediate types
    # When multiple fields reference the same base type (e.g., Cabin.doors and Vehicle.doors both reference [Door]),
    # we currently create intermediate types twice. Instead, maintain a cache:
    # expanded_types_cache: dict[str, GraphQLObjectType] = {}  # Maps base_type.name -> first_intermediate_type
    # Check cache before creating, reuse if exists, otherwise create once and cache.

    for parent_type, field_name in expandable_fields:
        original_field = parent_type.fields[field_name]
        base_type = cast(GraphQLObjectType, get_named_type(original_field.type))

        log.debug(f"Processing field '{field_name}' in type '{parent_type.name}' with base type '{base_type.name}'")

        unwrapped_type = original_field.type
        if is_non_null_type(unwrapped_type):
            unwrapped_type = unwrapped_type.of_type

        target_type = unwrapped_type.of_type if is_list_type(unwrapped_type) else unwrapped_type
        list_item_nullable = not is_non_null_type(target_type)

        instance_tag_object = cast(GraphQLObjectType, get_instance_tag_object(base_type, schema))
        instance_tag_dict = get_instance_tag_dict(instance_tag_object)
        instance_tag_types_to_remove.add(instance_tag_object.name)

        intermediate_types = _create_intermediate_types(base_type, instance_tag_dict, list_item_nullable)
        first_intermediate_type = intermediate_types[-1]

        type_case = get_case_for_element(ElementType.TYPE, ContextType.OBJECT, naming_config) if naming_config else None
        field_case = (
            get_case_for_element(ElementType.FIELD, ContextType.OBJECT, naming_config) if naming_config else None
        )

        for intermediate_type in intermediate_types:
            type_name = convert_name(intermediate_type.name, type_case) if type_case else intermediate_type.name
            intermediate_type.name = type_name
            new_types[type_name] = intermediate_type
            type_metadata[type_name] = TypeMetadata(source=None, is_intermediate_type=True)

        base_name = base_type.name if is_list_type(unwrapped_type) else field_name
        new_field_name = convert_name(base_name, field_case) if field_case else base_name

        resolved_names = [
            f"{new_field_name}.{'.'.join(instance_product)}"
            for instance_product in product(*instance_tag_dict.values())
        ]

        instances = list(instance_tag_dict.values())

        new_field = GraphQLField(
            type_=GraphQLNonNull(first_intermediate_type),
            description=original_field.description,
        )

        field_metadata[(parent_type.name, new_field_name)] = FieldMetadata(
            resolved_names=resolved_names,
            resolved_type=base_type.name,
            is_expanded=True,
            original_field=original_field,
            instances=instances,
        )

        del parent_type.fields[field_name]
        parent_type.fields[new_field_name] = new_field

        log.debug(f"Replaced field '{field_name}' with '{new_field_name}' in type '{parent_type.name}'")

        base_types_to_clean.add(base_type)

    all_types_to_remove = set(instance_tag_types_to_remove)
    all_instance_tag_types = get_all_objects_with_directive(get_all_object_types(schema), "instanceTag")
    all_types_to_remove.update(t.name for t in all_instance_tag_types)

    for base_type in base_types_to_clean:
        del base_type.fields["instanceTag"]
        log.debug(f"Removed 'instanceTag' field from type '{base_type.name}'")

    for type_name in all_types_to_remove:
        if type_name in schema.type_map:
            del schema.type_map[type_name]
            log.debug(f"Removed type '{type_name}' with @instanceTag directive from schema")

    for type_name, new_type in new_types.items():
        schema.type_map[type_name] = new_type

    log.info(f"Instance expansion complete. Created {len(new_types)} intermediate types")

    return schema, type_metadata, field_metadata



"""SHACL export route."""

from fastapi import APIRouter

from s2dm.api.config import COMMON_RESPONSES
from s2dm.api.errors import ResponseError, format_error_list
from s2dm.api.models.base import ApiResponse
from s2dm.api.models.shacl import ShaclExportRequest
from s2dm.api.services.response_service import execute_and_respond
from s2dm.api.services.schema_service import load_and_process_schema_wrapper
from s2dm.exporters.shacl import translate_to_shacl
from s2dm.exporters.utils.schema_loader import check_correct_schema

router = APIRouter(responses=COMMON_RESPONSES)


@router.post(
    "/shacl",
    response_model=ApiResponse,
    openapi_extra={"x-exporter-name": "SHACL", "x-cli-command-name": "shacl"},
)
def export_shacl(request: ShaclExportRequest) -> ApiResponse:
    """Export GraphQL schema to SHACL shapes."""

    def process_request() -> list[str]:
        annotated_schema, _ = load_and_process_schema_wrapper(
            schemas=request.schemas,
            naming_config_input=request.naming_config,
            selection_query_input=request.selection_query,
            root_type=request.root_type,
            expanded_instances=request.expanded_instances,
        )

        schema_errors = check_correct_schema(annotated_schema.schema)
        if schema_errors:
            raise ResponseError(format_error_list("Schema validation failed", schema_errors))

        graph = translate_to_shacl(
            annotated_schema,
            request.shapes_namespace,
            request.shapes_namespace_prefix,
            request.model_namespace,
            request.model_namespace_prefix,
        )

        result_str = graph.serialize(format=request.serialization_format)
        if isinstance(result_str, bytes):
            result_str = result_str.decode("utf-8")
        return [result_str]

    return execute_and_respond(executor=process_request, result_format=request.serialization_format)



from graphql import GraphQLField, GraphQLScalarType

from s2dm.exporters.utils.directive import get_directive_arguments, has_given_directive

INT32_MIN = -(2**31)
INT32_MAX = 2**31 - 1

GRAPHQL_SCALAR_TO_AVRO = {
    "String": "string",
    "Int": "int",
    "Float": "double",
    "Boolean": "boolean",
    "ID": "string",
    "Int8": "int",
    "UInt8": "int",
    "Int16": "int",
    "UInt16": "int",
    "UInt32": "long",
    "Int64": "long",
    "UInt64": "long",
}


def get_avro_scalar_type(scalar_type: GraphQLScalarType, field: GraphQLField | None = None) -> str:
    """Get Avro type for a GraphQL scalar, considering @range directive."""
    base_type = GRAPHQL_SCALAR_TO_AVRO.get(scalar_type.name, "string")

    if field and base_type in ("int", "long") and has_given_directive(field, "range"):
        range_args = get_directive_arguments(field, "range")

        for key in ("min", "max"):
            if key in range_args:
                val = range_args[key]
                if val < INT32_MIN or val > INT32_MAX:
                    return "long"

        return "int"

    return base_type



import click
import langcodes


def validate_language_tag(ctx: click.Context, param: click.Parameter, value: str) -> str:
    """Validate if a given string is compliant to BCP 47 language tag standard.

    Args:
        ctx: Click context
        param: Click parameter
        value: The language tag to validate

    Returns:
        The validated language tag string

    Raises:
        click.BadParameter: If the language tag is invalid
    """
    # Check for basic validity first
    if not value.strip():
        raise click.BadParameter("Language tag cannot be empty")

    # Check for valid BCP 47 format using langcodes
    try:
        if not langcodes.get(value).is_valid():
            raise click.BadParameter(f"'{value}' is not a valid BCP 47 language tag")
    except ValueError as e:
        # Handle langcodes.LanguageTagError (inherits from ValueError)
        raise click.BadParameter(f"'{value}' is not a valid BCP 47 language tag: {str(e)}") from None

    return value



from pydantic import Field

from s2dm.api.models.base import BaseExportRequest


class ShaclExportRequest(BaseExportRequest):
    """Request model for SHACL export endpoint."""

    serialization_format: str = Field(
        default="ttl",
        description="RDF serialization format",
        json_schema_extra={"x-cli-flag": "--serialization-format"},
    )
    shapes_namespace: str = Field(
        default="http://example.ns/shapes#",
        description="Namespace for SHACL shapes",
        json_schema_extra={"x-cli-flag": "--shapes-namespace"},
    )
    shapes_namespace_prefix: str = Field(
        default="shapes",
        description="Prefix for SHACL shapes namespace",
        json_schema_extra={"x-cli-flag": "--shapes-namespace-prefix"},
    )
    model_namespace: str = Field(
        default="http://example.ns/model#",
        description="Namespace for the data model",
        json_schema_extra={"x-cli-flag": "--model-namespace"},
    )
    model_namespace_prefix: str = Field(
        default="model",
        description="Prefix for the data model namespace",
        json_schema_extra={"x-cli-flag": "--model-namespace-prefix"},
    )



from collections.abc import Callable
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

import pytest
from ariadne import gql
from faker import Faker
from fastapi.testclient import TestClient
from graphql import (
    GraphQLNamedType,
    GraphQLSchema,
    build_schema,
)
from hypothesis import strategies as st
from hypothesis.strategies import composite

from s2dm.api.main import app
from s2dm.exporters.utils.extraction import get_all_named_types
from s2dm.exporters.utils.schema_loader import ensure_query
from s2dm.tools.graphql_inspector import locate_graphql_inspector
from s2dm.units.sync import UnitRow, _uri_to_enum_symbol

SCALAR_TYPES = ["String", "Int", "Float", "Boolean"]


class TestSchemaData:
    """Test schema data paths and constants."""

    TESTS_DATA_DIR: Path = Path(__file__).parent / "data"
    SCHEMA1: Path = TESTS_DATA_DIR / "schema1.graphql"
    SCHEMA2: Path = TESTS_DATA_DIR / "schema2.graphql"
    SCHEMA3: Path = TESTS_DATA_DIR / "schema3.graphql"
    UNITS_SCHEMA_PATH: Path = TESTS_DATA_DIR / "units"

    SAMPLE1_1 = TESTS_DATA_DIR / "schema1-1.graphql"
    SAMPLE1_2 = TESTS_DATA_DIR / "schema1-2.graphql"
    SAMPLE2_1 = TESTS_DATA_DIR / "schema2-1.graphql"
    SAMPLE2_2 = TESTS_DATA_DIR / "schema2-2.graphql"
    SAMPLE3 = TESTS_DATA_DIR / "schema3.graphql"

    VALID_QUERY = TESTS_DATA_DIR / "valid_query.graphql"
    INVALID_QUERY = TESTS_DATA_DIR / "invalid_query.graphql"
    SCHEMA1_QUERY = TESTS_DATA_DIR / "schema1_query.graphql"

    # Version bump test schemas
    BASE_SCHEMA = TESTS_DATA_DIR / "base.graphql"
    NO_CHANGE_SCHEMA = TESTS_DATA_DIR / "no-change.graphql"
    NON_BREAKING_SCHEMA = TESTS_DATA_DIR / "non-breaking.graphql"
    DANGEROUS_SCHEMA = TESTS_DATA_DIR / "dangerous.graphql"
    BREAKING_SCHEMA = TESTS_DATA_DIR / "breaking.graphql"


@pytest.fixture(scope="session")
def spec_directory() -> Path:
    """Return path to S2DM spec files."""
    return Path(__file__).parent.parent / "tests" / "data" / "spec"


def parsed_console_output() -> str:
    """Parse console output (placeholder function)."""
    return ""


@pytest.fixture(scope="module")
def schema_path(spec_directory: Path) -> list[Path]:
    assert TestSchemaData.SCHEMA1.exists(), f"Missing test file: {TestSchemaData.SCHEMA1}"
    assert TestSchemaData.UNITS_SCHEMA_PATH.exists(), f"Missing units folder: {TestSchemaData.UNITS_SCHEMA_PATH}"
    return [spec_directory, TestSchemaData.SCHEMA1, TestSchemaData.UNITS_SCHEMA_PATH]


@dataclass
class MockFieldData:
    name: str
    data_type: str = ""
    unit: str = ""
    allowed: list[str] = field(default_factory=list)
    minimum: int | None = None
    maximum: int | None = None

    @property
    def parent_name(self) -> str:
        return "Vehicle"

    @property
    def is_enum_field(self) -> bool:
        return bool(self.allowed)

    @property
    def enum_name(self) -> str:
        return f"{self.parent_name}_{self.name.capitalize()}_Enum"

    @property
    def enum_schema_str(self) -> str:
        return f"""
            enum {self.enum_name if self.is_enum_field else self.unit_enum_name} {{
                {" ".join(self.allowed or self.unit_allowed_values)}
            }}
        """

    @property
    def unit_enum_name(self) -> str:
        return f"{self.unit.capitalize()}Unit"

    @property
    def unit_default_value(self) -> str:
        return self.unit[:2]

    @property
    def unit_allowed_values(self) -> list[str]:
        return [self.unit_default_value]

    @classmethod
    def non_enum_field_data(cls, faker: Faker) -> "MockFieldData":
        """Generates a random non-enum field data with a random values
        e.g.
        length(unit: LengthUnit = MILLIMETER): Int
        """
        name = faker.unique.word().lower()
        data_type = faker.random_element(SCALAR_TYPES)
        unit = faker.unique.word()

        return cls(name, data_type, unit)

    @classmethod
    def enum_field_data(cls, faker: Faker) -> "MockFieldData":
        """Generates a random enum field data with a random values
        e.g.
        lowVoltageSystemState: Vehicle_LowVoltageSystemState_Enum
        """

        name = faker.unique.word().lower()
        num_values = faker.random_int(min=2, max=5)
        allowed = [faker.unique.word().upper() for _ in range(num_values)]

        return cls(name, data_type="string", allowed=allowed)

    def to_field_str(self) -> str:
        if self.is_enum_field:
            return f"{self.name}: {self.enum_name}"
        return f"{self.name}(unit: {self.unit_enum_name} = {self.unit_default_value}): {self.data_type}"


@composite
def mock_graphql_schema_strategy(
    draw: Callable[[st.SearchStrategy[Any]], Any],
) -> tuple[GraphQLSchema, list[MockFieldData]]:
    """Generate a random GraphQL schema with random types and fields."""
    faker = Faker()

    # Generate fields with a random name and a random type
    num_fields = draw(st.integers(min_value=2, max_value=5))
    fields = []
    enums = []
    for _ in range(num_fields):
        if draw(st.booleans()):
            # Generate an enum field
            field_data = MockFieldData.enum_field_data(faker)
        else:
            field_data = MockFieldData.non_enum_field_data(faker)

        enums.append(field_data.enum_schema_str)
        fields.append(field_data)

    schema_str = gql(
        f"""#graphql
        type Query {{
            vehicle: Vehicle
        }}

        {" ".join(enums)}

        type Vehicle {{
            {" ".join(map(lambda f: f.to_field_str(), fields))}
        }}
        """
    )

    return ensure_query(build_schema(gql(schema_str))), fields


@composite
def mock_named_types_strategy(
    draw: Callable[[st.SearchStrategy[Any]], Any],
) -> tuple[list[GraphQLNamedType], list[MockFieldData]]:
    schema, fields = draw(mock_graphql_schema_strategy())
    return get_all_named_types(schema), fields


# Constants for common test values
MOCK_QUDT_VERSION = "3.1.5"
MOCK_ENUM_CONTENT = "enum TestEnum { TEST1 }"

# QUDT constants for unit testing
QUDT_UNIT_BASE = "http://qudt.org/vocab/unit"
QUDT_QK_BASE = "http://qudt.org/vocab/quantitykind"

# Standard test paths for units sync
STANDARD_UNIT_PATHS = [
    "velocity/VelocityUnit.graphql",
    "mass/MassUnit.graphql",
    "length/LengthUnit.graphql",
]


@pytest.fixture
def mock_qudt_version() -> str:
    """Standard QUDT version for testing."""
    return MOCK_QUDT_VERSION


@pytest.fixture
def mock_get_latest_qudt_version(mock_qudt_version: str) -> Callable[[], str]:
    """Mock function for checking latest QUDT version."""

    def _mock() -> str:
        return mock_qudt_version

    return _mock


@pytest.fixture
def mock_sync_qudt_units() -> Callable[..., list[Path]]:
    """Mock function for sync_qudt_units with configurable behavior."""

    def _mock(
        units_root: Path,
        version: str | None = None,
        *,
        dry_run: bool = False,
        num_paths: int = 3,
        create_files: bool = True,
    ) -> list[Path]:
        # Generate test paths based on the number requested
        test_paths = [units_root / path for path in STANDARD_UNIT_PATHS[:num_paths]]

        # Simulate creating files only when not in dry-run mode and create_files is True
        if not dry_run and create_files:
            for path in test_paths:
                path.parent.mkdir(parents=True, exist_ok=True)
                path.write_text(MOCK_ENUM_CONTENT)

        return test_paths

    return _mock


@pytest.fixture
def units_sync_mocks(
    monkeypatch: pytest.MonkeyPatch,
    mock_sync_qudt_units: Callable[..., list[Path]],
    mock_get_latest_qudt_version: Callable[[], str],
    tmp_path: Path,
) -> tuple[Callable[..., list[Path]], Callable[[], str]]:
    """Fixture that applies all units sync mocks at once and isolates DEFAULT_QUDT_UNITS_DIR."""
    # Patch DEFAULT_QUDT_UNITS_DIR to use a temporary directory for test isolation
    test_units_dir = tmp_path / "test_units"
    monkeypatch.setattr("s2dm.cli.DEFAULT_QUDT_UNITS_DIR", test_units_dir)

    monkeypatch.setattr("s2dm.cli.sync_qudt_units", mock_sync_qudt_units)
    monkeypatch.setattr("s2dm.cli.get_latest_qudt_version", mock_get_latest_qudt_version)
    return mock_sync_qudt_units, mock_get_latest_qudt_version


def create_test_unit_row(
    unit_segment: str,
    qk_segment: str,
    label: str | None = None,
    ucum: str | None = None,
) -> "UnitRow":
    """Create a test UnitRow with minimal required data.

    Note: This imports UnitRow and _uri_to_enum_symbol locally to avoid circular imports.
    """
    unit_label = label or unit_segment.lower().replace("-", " ")
    symbol = _uri_to_enum_symbol(f"{QUDT_UNIT_BASE}/{unit_segment}")

    return UnitRow(
        unit_iri=f"{QUDT_UNIT_BASE}/{unit_segment}",
        unit_label=unit_label,
        quantity_kind_iri=f"{QUDT_QK_BASE}/{qk_segment}",
        quantity_kind_label=qk_segment,
        symbol=symbol,
        ucum_code=ucum or unit_segment.lower(),
    )


# Fixtures for ID Exporter tests
@pytest.fixture
def temp_output_paths(tmp_path: Path) -> dict[str, Path]:
    """Provide standard output paths for ID exporter tests."""
    return {
        "ids": tmp_path / "ids.json",
        "previous_ids": tmp_path / "previous_ids.json",
    }


@pytest.fixture
def schema_builder() -> Callable[[str], GraphQLSchema]:
    """Helper to build schemas consistently."""

    def _build(schema_str: str) -> GraphQLSchema:
        return ensure_query(build_schema(schema_str))

    return _build


# GraphQL Inspector fixtures and markers
@pytest.fixture(scope="session")
def inspector_path() -> Path | None:
    """Provide node_modules path for GraphQL Inspector tests.

    This fixture automatically locates the node_modules directory containing
    the @graphql-inspector/cli package. Tests that require GraphQL Inspector
    should use this fixture and be marked with @pytest.mark.graphql_inspector.

    Returns:
        Path to node_modules directory, or None if not found
    """
    path = locate_graphql_inspector()
    if path is None:
        pytest.skip("graphql-inspector not found. Run 'npm install' to install dependencies.")
    return path


def pytest_configure(config: pytest.Config) -> None:
    """Register custom markers for pytest."""
    config.addinivalue_line(
        "markers",
        (
            "graphql_inspector: mark test as requiring graphql-inspector CLI"
            " (can be skipped with -m 'not graphql_inspector')"
        ),
    )


@pytest.fixture(scope="module")
def test_client() -> TestClient:
    """FastAPI TestClient for API integration tests."""
    return TestClient(app)



import json
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pytest
from click.testing import CliRunner

from s2dm.cli import cli
from s2dm.tools.string import normalize_whitespace
from tests.conftest import TestSchemaData as TSD


@pytest.fixture(scope="session")
def units_directory() -> Path:
    """Return the test data units directory."""
    return TSD.UNITS_SCHEMA_PATH


@pytest.fixture(scope="module")
def runner() -> CliRunner:
    return CliRunner()


# Output files (will be created in a temp dir)
@pytest.fixture(scope="module")
def tmp_outputs(tmp_path_factory: pytest.TempPathFactory) -> Path:
    return tmp_path_factory.mktemp("e2e_outputs")


class ExpectedIds:
    """Expected spec history IDs for the test cases.

    The IDs are variant-based IDs in the format Concept/vN.
    - Initial IDs are v1.0 (from schema1-1/schema1-2)
    - Updated IDs are v2.0 (from schema2-1/schema2-2, after BREAKING changes)
    """

    VEHICLE_AVG_SPEED_ID = "Vehicle.averageSpeed/v1.0"  # schema1-1.graphql (Float)
    NEW_VEHICLE_AVG_SPEED_ID = "Vehicle.averageSpeed/v2.0"  # schema2-1.graphql (Int - changed from Float, BREAKING)
    PERSON_HEIGHT_ID = "Person.height/v1.0"  # schema1-2.graphql (Float)
    NEW_PERSON_HEIGHT_ID = "Person.height/v2.0"  # schema2-2.graphql (Int - changed from Float, BREAKING)


def contains_value(obj: dict[str, Any] | list[Any] | str, target: str) -> bool:
    """Helper function to recursively search dicts"""
    if isinstance(obj, dict):
        for v in obj.values():
            if contains_value(v, target):
                return True
    elif isinstance(obj, list):
        for item in obj:
            if contains_value(item, target):
                return True
    else:
        return obj == target
    return False


# ToDo(DA): please update this test to do proper asserts for the shacl exporter
def test_export_shacl(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "shacl.ttl"
    result = runner.invoke(
        cli,
        [
            "export",
            "shacl",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
            "-f",
            "ttl",
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()
    with open(out, encoding="utf-8") as f:
        content = f.read()

    assert "shapes:Vehicle" in content
    assert "shapes:Vehicle_ADAS_ObstacleDetection" in content


# ToDo(DA): please update this test to do proper asserts for the vspec exporter
def test_export_vspec(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "vspec.yaml"
    result = runner.invoke(
        cli,
        [
            "export",
            "vspec",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()
    with open(out, encoding="utf-8") as f:
        content = f.read()

    assert "Vehicle:" in content
    assert "Vehicle_ADAS_ObstacleDetection:" in content


def test_export_jsonschema(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "jsonschema.yaml"
    result = runner.invoke(
        cli,
        [
            "export",
            "jsonschema",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()
    with open(out, encoding="utf-8") as f:
        content = f.read()

    assert '"Vehicle"' in content
    assert '"Vehicle_ADAS_ObstacleDetection"' in content


def test_export_protobuf(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "schema.proto"
    result = runner.invoke(
        cli,
        [
            "export",
            "protobuf",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-q",
            str(TSD.SCHEMA1_QUERY),
            "-o",
            str(out),
            "-r",
            "Vehicle",
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()
    with open(out, encoding="utf-8") as f:
        content = f.read()

    assert "package" not in content

    assert "message Vehicle" in content
    assert "message Vehicle_ADAS" in content
    assert "message Vehicle_ADAS_ObstacleDetection" in content

    assert "message Vehicle_ADAS_ObstacleDetection_WarningType_Enum" in content
    assert "enum Enum" in content

    assert "optional float averageSpeed = 1;" in content
    assert "optional bool isEngaged = 1;" in content


def test_export_protobuf_flattened_naming(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    out = tmp_outputs / "schema.proto"
    result = runner.invoke(
        cli,
        [
            "export",
            "protobuf",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-q",
            str(TSD.SCHEMA1_QUERY),
            "-o",
            str(out),
            "-r",
            "Vehicle",
            "-f",
            "-p",
            "package.name",
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()
    with open(out, encoding="utf-8") as f:
        content = f.read()

    assert "package package.name;" in content

    assert "message Selection" in content

    assert "message Vehicle_ADAS_ObstacleDetection_WarningType_Enum" in content
    assert "enum Enum" in content

    assert 'optional float Vehicle_averageSpeed = 1 [(field_source) = "Vehicle"];' in content
    assert 'optional bool Vehicle_adas_abs_isEngaged = 3 [(field_source) = "Vehicle_ADAS_ABS"];' in content


def test_generate_skos_skeleton(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    out = tmp_outputs / "skos_skeleton.ttl"
    result = runner.invoke(
        cli,
        [
            "generate",
            "skos-skeleton",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()
    with open(out, encoding="utf-8") as f:
        content = f.read()

    assert "@prefix skos:" in content
    assert "skos:Concept" in content
    assert "skos:prefLabel" in content

    assert "Vehicle" in content
    assert "Vehicle_ADAS_ObstacleDetection" in content


@pytest.mark.parametrize(
    "schema_file,previous_file,expected_output",
    [
        ([TSD.NO_CHANGE_SCHEMA], [TSD.BASE_SCHEMA], "No version bump needed"),
        ([TSD.NON_BREAKING_SCHEMA], [TSD.BASE_SCHEMA], "Patch version bump needed"),
        ([TSD.DANGEROUS_SCHEMA], [TSD.BASE_SCHEMA], "Minor version bump needed"),
        ([TSD.BREAKING_SCHEMA], [TSD.BASE_SCHEMA], "Breaking changes detected - major version bump needed"),
        # Keep original test cases for backward compatibility
        ([TSD.SAMPLE1_1, TSD.SAMPLE1_2], [TSD.SAMPLE1_1, TSD.SAMPLE1_2], "No version bump needed"),
        (
            [TSD.SAMPLE1_1, TSD.SAMPLE1_2],
            [TSD.SAMPLE2_1, TSD.SAMPLE2_2],
            "Breaking changes detected - major version bump needed",
        ),
    ],
)
def test_check_version_bump(
    runner: CliRunner,
    schema_file: list[Path],
    previous_file: list[Path],
    expected_output: str,
    spec_directory: Path,
    units_directory: Path,
) -> None:
    result = runner.invoke(
        cli,
        ["check", "version-bump"]
        + ["-s", str(spec_directory)]
        + [item for schema in schema_file for item in ["-s", str(schema)]]
        + ["-s", str(units_directory)]
        + ["--previous", str(spec_directory)]
        + [item for previous in previous_file for item in ["--previous", str(previous)]]
        + ["--previous", str(units_directory)],
    )
    assert result.exit_code == 0, result.output
    # Replace all newlines and additional spaces with a single space with regex
    assert expected_output.lower() in normalize_whitespace(result.output).lower()


@pytest.mark.parametrize(
    "schema_file,previous_file,expected_type",
    [
        (TSD.NO_CHANGE_SCHEMA, TSD.BASE_SCHEMA, "none"),
        (TSD.NON_BREAKING_SCHEMA, TSD.BASE_SCHEMA, "patch"),
        (TSD.DANGEROUS_SCHEMA, TSD.BASE_SCHEMA, "minor"),
        (TSD.BREAKING_SCHEMA, TSD.BASE_SCHEMA, "major"),
    ],
)
def test_check_version_bump_output_type(
    runner: CliRunner, schema_file: Path, previous_file: Path, expected_type: str, spec_directory: Path
) -> None:
    result = runner.invoke(
        cli,
        [
            "check",
            "version-bump",
            "-s",
            str(spec_directory),
            "-s",
            str(schema_file),
            "--previous",
            str(spec_directory),
            "--previous",
            str(previous_file),
            "--output-type",
        ],
    )
    assert result.exit_code == 0, result.output
    # The output type should be the last line
    output_lines = result.output.strip().split("\n")
    assert output_lines[-1] == expected_type


# ToDo(DA): can you provide a negative example here?
@pytest.mark.parametrize(
    "input_file,expected_output",
    [
        ((TSD.SAMPLE1_1, TSD.SAMPLE1_2), "All constraints passed"),
    ],
)
def test_check_constraints(
    runner: CliRunner, input_file: tuple[Path, Path], expected_output: str, spec_directory: Path, units_directory: Path
) -> None:
    result = runner.invoke(
        cli,
        [
            "check",
            "constraints",
            "-s",
            str(spec_directory),
            "-s",
            str(input_file[0]),
            "-s",
            str(input_file[1]),
            "-s",
            str(units_directory),
        ],
    )
    assert expected_output.lower() in normalize_whitespace(result.output).lower()
    assert result.exit_code in (0, 1)


def test_validate_graphql(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "validate.json"
    result = runner.invoke(
        cli,
        [
            "validate",
            "graphql",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()
    with open(out) as f:
        file_content = f.read()
    assert "Vehicle" in file_content
    assert "Person" in file_content


@pytest.mark.parametrize(
    "schemas,val_schemas,expected_output",
    [
        ((TSD.SAMPLE1_1, TSD.SAMPLE1_2), (TSD.SAMPLE1_1, TSD.SAMPLE1_2), "No changes detected"),
        ((TSD.SAMPLE1_1, TSD.SAMPLE1_2), (TSD.SAMPLE2_1, TSD.SAMPLE2_2), "Detected"),
    ],
)
def test_diff_graphql(
    runner: CliRunner,
    tmp_outputs: Path,
    schemas: tuple[Path, Path],
    val_schemas: tuple[Path, Path],
    expected_output: str,
    spec_directory: Path,
    units_directory: Path,
) -> None:
    out = tmp_outputs / "diff.json"
    result = runner.invoke(
        cli,
        [
            "diff",
            "graphql",
            "-s",
            str(spec_directory),
            "-s",
            str(schemas[0]),
            "-s",
            str(schemas[1]),
            "-s",
            str(units_directory),
            "--val-schema",
            str(spec_directory),
            "--val-schema",
            str(val_schemas[0]),
            "--val-schema",
            str(val_schemas[1]),
            "--val-schema",
            str(units_directory),
            "-o",
            str(out),
        ],
    )
    # diff_graphql exits with code 1 if breaking changes detected, 0 otherwise
    # Both are valid outcomes, so we check if file was created
    assert result.exit_code in (0, 1), f"Command failed with exit code {result.exit_code}. Output: {result.output}"
    assert out.exists(), f"Output file {out} was not created. Exit code: {result.exit_code}, Output: {result.output}"
    with open(out) as f:
        diff_output = json.load(f)

    if expected_output == "No changes detected":
        assert len(diff_output) == 0
    elif expected_output == "Detected":
        assert len(diff_output) > 0


def test_registry_export_concept_uri(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    out = tmp_outputs / "concept_uris.json"
    result = runner.invoke(
        cli,
        [
            "registry",
            "concept-uri",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()
    with open(out) as f:
        data = json.load(f)

    assert isinstance(data, dict), "Expected JSON-LD output to be a dict."

    assert contains_value(
        data, "ns:Vehicle.averageSpeed"
    ), 'Expected value "ns:Vehicle.averageSpeed" not found in the concept URI output.'
    assert contains_value(
        data, "ns:Person.name"
    ), 'Expected value "ns:Person.name" not found in the concept URI output.'


def test_registry_export_id(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "ids.json"
    result = runner.invoke(
        cli,
        [
            "registry",
            "id",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
            "--version-tag",
            "v1.0.0",
        ],
    )
    assert result.exit_code == 0, f"Expected exit code 0, but got {result.exit_code}."
    out = out.with_name(f"{out.stem}_v1.0.0{out.suffix}")
    assert out.exists(), f"Expected output file {out} not found."
    with open(out) as f:
        data = json.load(f)
    # New format has 'concepts' key with concept names as keys
    concepts = data["concepts"]
    assert any(
        "Vehicle.averageSpeed" in k for k in concepts
    ), "Expected 'Vehicle.averageSpeed' not found in the output."
    assert any("Person.name" in k for k in concepts), "Expected 'Person.name' not found in the output."


def test_registry_init(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "spec_history.json"
    result = runner.invoke(
        cli,
        [
            "registry",
            "init",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
            "--version-tag",
            "v1.0.0",
        ],
    )
    assert result.exit_code == 0, result.output
    out = out.with_name(f"{out.stem}_v1.0.0{out.suffix}")
    assert out.exists()
    with open(out) as f:
        data = json.load(f)

    found_vehicle = False
    found_person = False
    entries = data if isinstance(data, list) else data.get("@graph") or data.get("items") or []
    for entry in entries:
        if isinstance(entry, dict):
            if entry.get("@id") == "ns:Vehicle.averageSpeed":
                spec_history = entry.get("specHistory", [])
                if (
                    spec_history
                    and isinstance(spec_history, list)
                    and isinstance(spec_history[0], dict)
                    and spec_history[0].get("@id") == ExpectedIds.VEHICLE_AVG_SPEED_ID
                ):
                    found_vehicle = True
            elif entry.get("@id") == "ns:Person.height":
                spec_history = entry.get("specHistory", [])
                if (
                    spec_history
                    and isinstance(spec_history, list)
                    and isinstance(spec_history[0], dict)
                    and spec_history[0].get("@id") == ExpectedIds.PERSON_HEIGHT_ID
                ):
                    found_person = True
        if found_vehicle and found_person:
            break
    assert found_vehicle, (
        'Expected entry with "@id": "ns:Vehicle.averageSpeed" and specHistory id'
        + f'"{ExpectedIds.VEHICLE_AVG_SPEED_ID}" not found.'
    )
    assert (
        found_person
    ), f'Expected entry with "@id": "ns:Person.height" and specHistory id "{ExpectedIds.PERSON_HEIGHT_ID}" not found.'


def test_registry_update(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "spec_history_update.json"
    version_tag_initial = "v1.0.0"
    version_tag_updated = "v1.1.0"
    # First, create a spec history file
    init_out = tmp_outputs / "spec_history.json"
    init_result = runner.invoke(
        cli,
        [
            "registry",
            "init",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(init_out),
            "--version-tag",
            version_tag_initial,
        ],
    )
    assert init_result.exit_code == 0, f"registry init failed: {init_result.output}"
    init_out = init_out.with_name(f"{init_out.stem}_{version_tag_initial}{init_out.suffix}")
    assert init_out.exists(), f"Init output file {init_out} was not created. Output: {init_result.output}"

    # Generate diff between old and new schemas
    diff_file = tmp_outputs / "diff.json"
    diff_result = runner.invoke(
        cli,
        [
            "diff",
            "graphql",
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "--val-schema",
            str(TSD.SAMPLE2_1),
            "--val-schema",
            str(TSD.SAMPLE2_2),
            "--val-schema",
            str(units_directory),
            "-o",
            str(diff_file),
        ],
    )
    # diff_graphql exits with code 1 if breaking changes detected, 0 otherwise
    assert diff_result.exit_code in (0, 1), f"diff graphql failed: {diff_result.output}"
    assert diff_file.exists(), f"Diff file {diff_file} was not created. Output: {diff_result.output}"

    # Get previous variant IDs file (created during init)
    previous_ids = init_out.parent / f"variant_ids_{version_tag_initial}.json"
    assert previous_ids.exists(), f"Previous IDs file {previous_ids} was not created during init"

    update_result = runner.invoke(
        cli,
        [
            "registry",
            "update",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE2_1),
            "-s",
            str(TSD.SAMPLE2_2),
            "-s",
            str(units_directory),
            "-sh",
            str(init_out),
            "-o",
            str(out),
            "--previous-ids",
            str(previous_ids),
            "--diff-file",
            str(diff_file),
            "--version-tag",
            version_tag_updated,
        ],
    )
    assert update_result.exit_code == 0, f"registry update failed: {update_result.output}"
    out = out.with_name(f"{out.stem}_{version_tag_updated}{out.suffix}")
    assert out.exists(), f"Update output file {out} was not created. Output: {update_result.output}"
    with open(out) as f:
        data = json.load(f)
    found_vehicle_old = False
    found_vehicle_new = False
    found_person_old = False
    found_person_new = False
    entries = data if isinstance(data, list) else data.get("@graph") or data.get("items") or []
    for entry in entries:
        if isinstance(entry, dict):
            if entry.get("@id") == "ns:Vehicle.averageSpeed":
                spec_history = entry.get("specHistory", [])
                ids = [h.get("@id") for h in spec_history if isinstance(h, dict)]
                if ExpectedIds.VEHICLE_AVG_SPEED_ID in ids:
                    found_vehicle_old = True
                if ExpectedIds.NEW_VEHICLE_AVG_SPEED_ID in ids:
                    found_vehicle_new = True
            elif entry.get("@id") == "ns:Person.height":
                spec_history = entry.get("specHistory", [])
                ids = [h.get("@id") for h in spec_history if isinstance(h, dict)]
                if ExpectedIds.PERSON_HEIGHT_ID in ids:
                    found_person_old = True
                if ExpectedIds.NEW_PERSON_HEIGHT_ID in ids:
                    found_person_new = True
        if found_vehicle_old and found_vehicle_new and found_person_old and found_person_new:
            break
    assert (
        found_vehicle_old
    ), f'Expected old specHistory id "{ExpectedIds.VEHICLE_AVG_SPEED_ID}" for Vehicle.averageSpeed not found.'
    assert (
        found_vehicle_new
    ), f'Expected new specHistory id "{ExpectedIds.NEW_VEHICLE_AVG_SPEED_ID}" for Vehicle.averageSpeed not found.'
    assert (
        found_person_old
    ), f'Expected old specHistory id "{ExpectedIds.PERSON_HEIGHT_ID}" for Person.height not found.'
    assert (
        found_person_new
    ), f'Expected new specHistory id "{ExpectedIds.NEW_PERSON_HEIGHT_ID}" for Person.height not found.'


@pytest.mark.parametrize(
    "search_term,expected_output",
    [
        ("Vehicle", "Vehicle"),
        ("averageSpeed", "Vehicle: ['averageSpeed']"),
        ("id", "Vehicle: ['id']"),
        ("NonExistentType", "No matches found"),
        ("nonExistentField", "No matches found"),
    ],
)
def test_search_graphql(
    runner: CliRunner, search_term: str, expected_output: str, spec_directory: Path, units_directory: Path
) -> None:
    result = runner.invoke(
        cli,
        [
            "search",
            "graphql",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-t",
            search_term,
            "--exact",
        ],
    )
    assert result.exit_code == 0, result.output
    assert expected_output.lower() in normalize_whitespace(result.output).lower()


def test_search_skos(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    skos_file = tmp_outputs / "test_skos.ttl"
    result = runner.invoke(
        cli,
        [
            "generate",
            "skos-skeleton",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(skos_file),
        ],
    )
    assert result.exit_code == 0, result.output
    assert skos_file.exists()

    result = runner.invoke(cli, ["search", "skos", "-f", str(skos_file), "-t", "Vehicle"])
    assert result.exit_code == 0, result.output
    assert "Vehicle" in normalize_whitespace(result.output)

    result = runner.invoke(
        cli,
        ["search", "skos", "-f", str(skos_file), "-t", "vehicle", "--case-insensitive"],
    )
    assert result.exit_code == 0, result.output
    assert "Vehicle" in normalize_whitespace(result.output)

    result = runner.invoke(cli, ["search", "skos", "-f", str(skos_file), "-t", "NonExistentConcept"])
    assert result.exit_code == 0, result.output
    assert "No matches found" in normalize_whitespace(result.output)


@pytest.mark.parametrize(
    "search_term,expected_returncode,expected_output",
    [("Vehicle", 0, "Vehicle"), ("Seat", 1, "Type 'Seat' doesn't exist")],
)
def test_similar_graphql(
    runner: CliRunner,
    tmp_outputs: Path,
    search_term: str,
    expected_returncode: int,
    expected_output: str,
    spec_directory: Path,
    units_directory: Path,
) -> None:
    out = tmp_outputs / "similar.json"
    result = runner.invoke(
        cli,
        [
            "similar",
            "graphql",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-k",
            search_term,
            "-o",
            str(out),
        ],
    )
    assert expected_returncode == result.exit_code, result.output
    assert expected_output in normalize_whitespace(result.output)
    assert out.exists()


def test_compose_graphql(runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path) -> None:
    out = tmp_outputs / "composed.graphql"
    result = runner.invoke(
        cli,
        [
            "compose",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()

    composed_content = out.read_text()
    assert "type Vehicle" in composed_content
    assert "type Vehicle_ADAS" in composed_content
    assert "type Vehicle_ADAS_ObstacleDetection" in composed_content
    assert "enum AccelerationUnit" in composed_content
    assert "directive @range" in composed_content

    assert "Successfully composed schema" in normalize_whitespace(result.output)


def test_compose_graphql_with_root_type(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    out = tmp_outputs / "composed_filtered.graphql"
    result = runner.invoke(
        cli,
        [
            "compose",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-r",
            "Vehicle",
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0, result.output
    assert out.exists()

    composed_content = out.read_text()
    assert "type Query" in composed_content

    assert "type Vehicle" in composed_content
    assert "type Vehicle_ADAS" in composed_content
    assert "Successfully composed schema with root type 'Vehicle'" in normalize_whitespace(result.output)

    assert "type Person" not in composed_content


def test_compose_graphql_with_root_type_nonexistent(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    out = tmp_outputs / "composed_error.graphql"
    result = runner.invoke(
        cli,
        [
            "compose",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-r",
            "NonExistentType",
            "-o",
            str(out),
        ],
    )

    assert result.exit_code == 1
    assert not out.exists()
    assert "Root type 'NonExistentType' not found in schema" in normalize_whitespace(result.output)


def test_compose_graphql_root_type_filters_unreferenced_types(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    out = tmp_outputs / "composed_filtered.graphql"
    result = runner.invoke(
        cli,
        [
            "compose",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-r",
            "Vehicle_ADAS",
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0

    composed_content = out.read_text()

    assert "type Vehicle_ADAS" in composed_content
    assert "type Vehicle_ADAS_ABS" in composed_content

    assert "type Vehicle_Body" not in composed_content
    assert "type Vehicle_Occupant" not in composed_content
    assert "type InCabinArea2x2" not in composed_content


def test_compose_preserves_custom_directives(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    """Test that compose preserves all types of custom directives and formatting."""
    out = tmp_outputs / "directive_preservation_test.graphql"
    result = runner.invoke(
        cli,
        [
            "compose",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
        ],
    )

    assert result.exit_code == 0, result.output
    assert out.exists()

    composed_content = out.read_text()

    assert "directive @range(min: Float, max: Float) on FIELD_DEFINITION" in composed_content
    assert "directive @cardinality(min: Int, max: Int) on FIELD_DEFINITION" in composed_content
    assert "directive @noDuplicates on FIELD_DEFINITION" in composed_content
    assert "directive @instanceTag on OBJECT" in composed_content

    assert "type Vehicle" in composed_content
    assert "type Vehicle_ADAS_ObstacleDetection" in composed_content


def test_compose_adds_reference_directives(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    """Test that compose adds @reference directives to track source files."""
    out = tmp_outputs / "reference_directives_test.graphql"
    result = runner.invoke(
        cli,
        [
            "compose",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
            "-o",
            str(out),
        ],
    )

    assert result.exit_code == 0, result.output
    assert out.exists()

    composed_content = out.read_text()

    assert 'type Vehicle @reference(source: "schema1-1.graphql")' in composed_content
    assert 'type Vehicle_ADAS @reference(source: "schema1-1.graphql")' in composed_content
    assert 'enum Vehicle_ADAS_ActiveAutonomyLevel_Enum @reference(source: "schema1-1.graphql")' in composed_content

    assert 'type Vehicle_ADAS_ObstacleDetection @reference(source: "schema1-2.graphql")' in composed_content
    assert 'type Vehicle_Occupant @reference(source: "schema1-2.graphql")' in composed_content
    assert 'type Person @reference(uri: "http://example.com")' in composed_content

    assert 'type InCabinArea2x2 @instanceTag @reference(source: "common_types.graphql")' in composed_content


def test_compose_reference_directive_placement_after_other_directives(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path
) -> None:
    """Test that @reference directives are placed correctly after implements clause."""
    # Create test schema with implements clause
    test_schema_dir = tmp_outputs / "test_schemas"
    test_schema_dir.mkdir()

    schema_file = test_schema_dir / "test.graphql"
    schema_file.write_text("""
scalar DateTime

interface Node {
  id: ID!
}

interface Timestamped {
  createdAt: String
  updatedAt: String
}

type User implements Node {
  id: ID!
  name: String
}

type Admin implements Node & Timestamped @instanceTag {
  id: ID!
  role: String
  createdAt: String
  updatedAt: String
}

union Person = User | Admin
""")

    out = tmp_outputs / "implements_test.graphql"
    result = runner.invoke(cli, ["compose", "-s", str(spec_directory), "-s", str(schema_file), "-o", str(out)])

    assert result.exit_code == 0, result.output
    composed_content = out.read_text()

    assert 'type User implements Node @reference(source: "test.graphql")' in composed_content
    assert (
        'type Admin implements Node & Timestamped @instanceTag @reference(source: "test.graphql")' in composed_content
    )

    assert 'scalar DateTime @reference(source: "test.graphql")' in composed_content
    assert 'union Person @reference(source: "test.graphql")' in composed_content


def test_compose_with_invalid_selection_query(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    out = tmp_outputs / "composed_invalid_query.graphql"
    result = runner.invoke(
        cli,
        [
            "compose",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE2_1),
            "-s",
            str(TSD.SAMPLE2_2),
            "-s",
            str(units_directory),
            "-q",
            str(TSD.INVALID_QUERY),
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 1


def test_compose_with_valid_selection_query_prunes_schema(
    runner: CliRunner, tmp_outputs: Path, spec_directory: Path, units_directory: Path
) -> None:
    out = tmp_outputs / "composed_pruned.graphql"
    result = runner.invoke(
        cli,
        [
            "compose",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE2_1),
            "-s",
            str(TSD.SAMPLE2_2),
            "-s",
            str(units_directory),
            "-q",
            str(TSD.VALID_QUERY),
            "-o",
            str(out),
        ],
    )
    assert result.exit_code == 0

    composed_content = out.read_text()

    assert "type Vehicle" in composed_content
    assert "type Vehicle_ADAS" in composed_content
    assert "type Vehicle_ADAS_ABS" in composed_content
    assert "enum Vehicle_LowVoltageSystemState_Enum" in composed_content
    assert "enum VelocityUnit" in composed_content

    assert "type Person" not in composed_content
    assert "type Vehicle_Body" not in composed_content
    assert "type Vehicle_Occupant" not in composed_content
    assert "enum Vehicle_ADAS_ActiveAutonomyLevel_Enum" not in composed_content

    assert "directive @reference" in composed_content

    assert "directive @range" not in composed_content
    assert "directive @cardinality" not in composed_content
    assert "directive @noDuplicates" not in composed_content
    assert "directive @instanceTag" not in composed_content
    assert "directive @metadata" not in composed_content


# ToDo(DA): needs refactoring after final decision how stats will work
def test_stats_graphql(runner: CliRunner, spec_directory: Path, units_directory: Path) -> None:
    result = runner.invoke(
        cli,
        [
            "stats",
            "graphql",
            "-s",
            str(spec_directory),
            "-s",
            str(TSD.SAMPLE1_1),
            "-s",
            str(TSD.SAMPLE1_2),
            "-s",
            str(units_directory),
        ],
    )
    print(f"{result.output=}")
    assert result.exit_code == 0, normalize_whitespace(result.output)
    assert '"UInt32": 1' in normalize_whitespace(result.output)


def test_units_sync_cli(
    runner: CliRunner,
    units_sync_mocks: tuple[Callable[..., list[Path]], Callable[[], str]],
) -> None:
    """Test that the units sync CLI command works end-to-end."""
    # Test basic CLI functionality
    result = runner.invoke(cli, ["units", "sync"])
    assert result.exit_code == 0, result.output
    assert "Generated" in normalize_whitespace(result.output) or "enum files" in normalize_whitespace(result.output)

    # Test that --dry-run flag is accepted (detailed behavior tested in unit tests)
    result = runner.invoke(cli, ["units", "sync", "--dry-run"])
    assert result.exit_code == 0, result.output
    assert (
        "Would generate" in normalize_whitespace(result.output) or "dry" in normalize_whitespace(result.output).lower()
    )



from graphql import DocumentNode

from s2dm import log
from s2dm.exporters.utils.annotated_schema import AnnotatedSchema

from .transformer import ProtobufTransformer


def transform(
    annotated_schema: AnnotatedSchema,
    selection_query: DocumentNode,
    package_name: str | None = None,
    flatten_root_types: list[str] | None = None,
) -> str:
    """
    Transform an annotated GraphQL schema to Protocol Buffers format.

    Args:
        annotated_schema: The annotated GraphQL schema object
        selection_query: Required selection query document to determine root-level types
        package_name: Optional package name for the .proto file
        flatten_root_types: Optional list of root type names for flatten mode

    Returns:
        str: Protocol Buffers representation as a string

    Raises:
        ValueError: If selection_query is not provided
    """
    log.info(f"Transforming GraphQL schema to Protobuf with {len(annotated_schema.schema.type_map)} types")

    transformer = ProtobufTransformer(annotated_schema, selection_query, package_name, flatten_root_types)
    proto_content = transformer.transform()

    log.info("Successfully converted GraphQL schema to Protobuf")

    return proto_content


def translate_to_protobuf(
    annotated_schema: AnnotatedSchema,
    selection_query: DocumentNode,
    package_name: str | None = None,
    flatten_root_types: list[str] | None = None,
) -> str:
    """
    Translate a GraphQL schema to Protocol Buffers format.

    Args:
        annotated_schema: The annotated GraphQL schema object
        selection_query: Required selection query document to determine root-level types
        package_name: Optional package name for the .proto file
        flatten_root_types: Optional list of root type names for flatten mode

    Returns:
        str: Protocol Buffers (.proto) representation as a string

    Raises:
        ValueError: If selection_query is not provided
    """
    return transform(annotated_schema, selection_query, package_name, flatten_root_types)



"""Avro export routes."""

from typing import cast

from fastapi import APIRouter
from graphql import DocumentNode

from s2dm.api.config import COMMON_RESPONSES
from s2dm.api.errors import ResponseError, format_error_list
from s2dm.api.models.avro import AvroProtocolExportRequest, AvroSchemaExportRequest
from s2dm.api.models.base import ApiResponse
from s2dm.api.services.response_service import execute_and_respond
from s2dm.api.services.schema_service import load_and_process_schema_wrapper
from s2dm.exporters.avro import translate_to_avro_protocol, translate_to_avro_schema
from s2dm.exporters.utils.schema_loader import check_correct_schema

router = APIRouter(responses=COMMON_RESPONSES)


@router.post(
    "/avro/schema",
    response_model=ApiResponse,
    openapi_extra={"x-exporter-name": "Avro Schema", "x-cli-command-name": "avro schema"},
)
def export_avro_schema(request: AvroSchemaExportRequest) -> ApiResponse:
    """Export GraphQL schema to Avro Schema."""

    def process_request() -> list[str]:
        annotated_schema, query_document = load_and_process_schema_wrapper(
            schemas=request.schemas,
            naming_config_input=request.naming_config,
            selection_query_input=request.selection_query,
            root_type=request.root_type,
            expanded_instances=request.expanded_instances,
        )

        schema_errors = check_correct_schema(annotated_schema.schema)
        if schema_errors:
            raise ResponseError(format_error_list("Schema validation failed", schema_errors))

        avro_schema = translate_to_avro_schema(annotated_schema, request.namespace, cast(DocumentNode, query_document))
        return [avro_schema]

    return execute_and_respond(executor=process_request, result_format="avsc")


@router.post(
    "/avro/protocol",
    response_model=ApiResponse,
    openapi_extra={"x-exporter-name": "Avro Protocol", "x-cli-command-name": "avro protocol"},
)
def export_avro_protocol(request: AvroProtocolExportRequest) -> ApiResponse:
    """Export GraphQL schema to Avro Protocol (IDL)."""

    def process_request() -> list[str]:
        annotated_schema, _ = load_and_process_schema_wrapper(
            schemas=request.schemas,
            naming_config_input=request.naming_config,
            selection_query_input=request.selection_query,
            root_type=request.root_type,
            expanded_instances=request.expanded_instances,
        )

        schema_errors = check_correct_schema(annotated_schema.schema)
        if schema_errors:
            raise ResponseError(format_error_list("Schema validation failed", schema_errors))

        protocols = translate_to_avro_protocol(annotated_schema, request.namespace, request.strict)
        return list(protocols.values())

    return execute_and_respond(executor=process_request, result_format="avdl")



import re
from pathlib import Path
from typing import Any

from s2dm import log
from s2dm.registry.concept_uris import ConceptUriModel
from s2dm.registry.spec_history import (
    SpecHistoryModel,
    convert_concept_uri_to_spec_history,
    load_json_file,
    save_spec_history,
    update_spec_history_from_concept_uris,
)


class SpecHistoryExporter:
    def __init__(
        self,
        schema_content: str,
        output: Path | None,
        history_dir: Path,
    ):
        """
        Args:
            concept_uri: Path to the concept URI JSON-LD file
            ids: Path to the IDs JSON file
            schemas: List of paths to GraphQL schema files to extract type definitions
            init: Whether to initialize a new spec history (True) or update (False)
            spec_history: Path to an existing spec history JSON-LD file (for updates)
            output: Path to the output spec history JSON-LD file
            history_dir: Directory to store type history files
        """
        self.schema_content = schema_content
        self.output = output
        self.history_dir = history_dir

    def _extract_type_definition(self, type_name: str) -> str | None:
        """
        Extract a GraphQL type definition from the schema file.

        Args:
            type_name: Name of the type to extract

        Returns:
            The complete type definition as a string, or None if not found
        """
        pattern = rf"(type|enum)\s+{re.escape(type_name)}\s*{{[^{{}}]*}}"
        match = re.search(pattern, self.schema_content, re.DOTALL)
        if match:
            return match.group(0)
        return None

    def init_spec_history_model(
        self,
        concept_uris: dict[str, Any],
        variant_ids: dict[str, Any],
        concept_uri_model: ConceptUriModel,
        version_tag: str | None = None,
    ) -> SpecHistoryModel:
        """
        Generate a first spec history registry to track changes in concept realizations.

        This method tracks the history of realization IDs for each concept over time.
        It can either initialize a new spec history or update an existing one.

        For initialization (self.init == True), provide:
        - A concept URI file (self.concept_uri)
        - An IDs file (self.ids)
        - An output path (self.output)

        For updates (self.init == False), also provide:
        - An existing spec history file (self.spec_history)

        For saving type definitions:
        - A GraphQL schema file (self.schemas)
        - Optionally, a directory to store type history (self.history_dir, default: "./history")

        Args:
            concept_uris: Dictionary of concept URIs
            variant_ids: Dictionary mapping concept names to variant IDs
            concept_uri_model: The concept URI model
            version_tag: The version tag for this spec history (e.g., "v1.0.0")
        """
        log.debug(f"Initializing new spec history from {concept_uris} and {variant_ids}")
        result = convert_concept_uri_to_spec_history(concept_uri_model, variant_ids, version_tag)
        if self.output:
            save_spec_history(result, self.output)
            log.info(f"Spec history initialized and saved to {self.output}")
        else:
            log.debug(result.model_dump(by_alias=True))
        self.process_type_definitions(list(variant_ids.keys()), [], variant_ids, self.history_dir)
        return result

    def update_spec_history_model(
        self,
        concept_uris: dict[str, Any],
        variant_ids: dict[str, Any],
        concept_uri_model: ConceptUriModel,
        spec_history_path: Path,
        version_tag: str | None = None,
    ) -> SpecHistoryModel:
        """
        Update a spec history registry to track changes in concept realizations.

        This method tracks the history of realization IDs for each concept over time.
        It can either initialize a new spec history or update an existing one.

        For update provide:
        - A concept URI file (self.concept_uri)
        - A variant IDs file (self.ids)
        - An output path (self.output)
        - An existing spec history file (self.spec_history)

        For saving type definitions:
        - A GraphQL schema file (self.schemas)
        - Optionally, a directory to store type history (self.history_dir, default: "./history")

        Args:
            concept_uris: Dictionary of concept URIs
            variant_ids: Dictionary mapping concept names to variant IDs
            concept_uri_model: The concept URI model
            spec_history_path: Path to the existing spec history file
            version_tag: The version tag for new/updated entries (e.g., "v1.0.0")
        """
        log.info(f"Updating spec history {spec_history_path} with {concept_uris} and {variant_ids}")
        existing_history_data = load_json_file(spec_history_path)
        existing_history = SpecHistoryModel.model_validate(existing_history_data)
        new_concepts, updated_ids = update_spec_history_from_concept_uris(
            existing_history, concept_uri_model, variant_ids, version_tag
        )
        if new_concepts:
            log.info(f"Added {len(new_concepts)} new concepts:")
            for new_concept in new_concepts:
                log.info(f"  {new_concept}")
        if updated_ids:
            log.info(f"Updated IDs for {len(updated_ids)} concepts:")
            for updated_id in updated_ids:
                log.info(f"  {updated_id}")
        if new_concepts or updated_ids:
            self.process_type_definitions(new_concepts, updated_ids, variant_ids, self.history_dir)
        if self.output:
            save_spec_history(existing_history, self.output)
            log.info(f"Updated spec history saved to {self.output}")
        else:
            log.info(existing_history.model_dump(by_alias=True))

        return existing_history

    def process_type_definitions(
        self,
        new_concepts: list[str],
        updated_ids: list[str],
        variant_ids: dict[str, str],
        history_dir: Path,
    ) -> None:
        """
        Process and save type definitions for new or updated concepts.

        Args:
            new_concepts: List of new concept names
            updated_ids: List of updated concept names
            variant_ids: Dictionary mapping concept names to their variant IDs
            history_dir: Directory to save type definitions in
        """
        log.info(f"Processing type definitions for {len(new_concepts)} new and {len(updated_ids)} updated concepts")
        concepts_to_process = new_concepts + updated_ids
        for concept_name in concepts_to_process:
            if concept_name not in variant_ids:
                log.warning(f"No variant ID found for concept {concept_name}, skipping")
                continue
            parent_type = concept_name.split(".")[0] if "." in concept_name else concept_name
            id_value = variant_ids[concept_name]
            type_def = self._extract_type_definition(parent_type)
            if type_def:
                self.save_type_definition(id_value, parent_type, type_def, history_dir)
            else:
                log.debug(f"Could not extract type definition for {parent_type}")

    def run(
        self, concept_uris_path: Path, variant_ids_path: Path, init: bool, spec_history_path: Path | None = None
    ) -> SpecHistoryModel:
        """
        Generate or update a spec history registry to track changes in concept realizations.

        This method tracks the history of realization IDs for each concept over time.
        It can either initialize a new spec history or update an existing one.

        For initialization (init == True), provide:
        - A concept URI file (self.concept_uri)
        - An IDs file (self.ids)
        - An output path (self.output)

        For updates (init == False), also provide:
        - An existing spec history file (self.spec_history)

        For saving type definitions:
        - A GraphQL schema file (self.schemas)
        - Optionally, a directory to store type history (self.history_dir, default: "./history")
        """
        # Load the concept URIs and IDs
        concept_uris_data = load_json_file(concept_uris_path)
        variant_ids = load_json_file(variant_ids_path)
        concept_uri_model = ConceptUriModel.model_validate(concept_uris_data)

        if init:
            return self.init_spec_history_model(concept_uris_data, variant_ids, concept_uri_model)

        if not spec_history_path:
            raise ValueError("spec history is required when using update")

        return self.update_spec_history_model(concept_uris_data, variant_ids, concept_uri_model, spec_history_path)

    @staticmethod
    def generate_history_filename(id_value: str) -> str:
        """Generate a filename for a type definition history file.

        Args:
            id_value: The ID value for the concept (variant ID format: Concept/vN)

        Returns:
            Filename in the format <sanitized_id>.graphql
            The ID is sanitized for filesystem compatibility by replacing '/' with '_'.
        """
        # Sanitize ID for filesystem: replace '/' with '_'
        sanitized_id = id_value.replace("/", "_")
        return f"{sanitized_id}.graphql"

    @staticmethod
    def save_type_definition(
        id_value: str,
        parent_type: str,
        type_def: str,
        history_dir: Path,
    ) -> None:
        """Save a type definition to a file in the history directory.

        Args:
            id_value: The ID value for the concept (variant ID format: Concept/vN)
            parent_type: The parent type name
            type_def: The complete type definition as a string
            history_dir: Directory to save the file in
        """
        history_dir.mkdir(parents=True, exist_ok=True)
        filename = SpecHistoryExporter.generate_history_filename(id_value)
        file_path: Path = history_dir / filename

        with open(file_path, "w") as f:
            f.write(type_def)
        log.debug(f"Saved type definition for {parent_type} to {file_path}")



from dataclasses import dataclass, field

from graphql import GraphQLField, GraphQLSchema


@dataclass
class TypeMetadata:
    source: str | None
    is_intermediate_type: bool


@dataclass
class FieldMetadata:
    resolved_names: list[str]
    resolved_type: str
    is_expanded: bool
    original_field: GraphQLField | None = None
    instances: list[list[str]] = field(default_factory=list)


@dataclass
class AnnotatedSchema:
    schema: GraphQLSchema
    type_metadata: dict[str, TypeMetadata] = field(default_factory=dict)
    field_metadata: dict[tuple[str, str], FieldMetadata] = field(default_factory=dict)



from graphql import (
    GraphQLEnumType,
    GraphQLList,
    GraphQLNamedType,
    GraphQLObjectType,
    GraphQLScalarType,
    get_named_type,
    is_scalar_type,
)

from s2dm import log
from s2dm.concept.models import Concepts, FieldMetadata
from s2dm.exporters.utils.graphql_type import is_id_type, is_introspection_or_root_type


def iter_all_concepts(named_types: list[GraphQLNamedType]) -> Concepts:
    """Extract all concepts from GraphQL named types with enhanced metadata.

    This function extracts all concepts and captures additional GraphQL
    field definitions for enhanced functionality (like SKOS generation),
    while maintaining backward compatibility.

    Args:
        named_types: List of GraphQL named types to process

    Returns:
        Concepts object containing all extracted concepts plus field metadata
    """
    concepts = Concepts()
    for named_type in named_types:
        if is_introspection_or_root_type(named_type.name):
            continue

        if isinstance(named_type, GraphQLEnumType):
            log.debug(f"Processing enum: {named_type.name}")
            concepts.enums.append(named_type.name)

        elif isinstance(named_type, GraphQLObjectType):
            log.debug(f"Processing object: {named_type.name}")
            # Get the ID of all fields in the object
            for field_name, field in named_type.fields.items():
                # Check if field's type is ID scalar - skip ID fields
                field_type = get_named_type(field.type)
                if (
                    is_scalar_type(field_type)
                    and isinstance(field_type, GraphQLScalarType)
                    and is_id_type(field_type.name)
                ):
                    continue

                field_fqn = f"{named_type.name}.{field_name}"

                if isinstance(field.type, GraphQLObjectType):
                    # field uses the object type
                    concepts.nested_objects[field_fqn] = field.type.name
                elif isinstance(field.type, GraphQLList):
                    # field uses a list of the object type
                    internal_type = field.type
                    while hasattr(internal_type, "of_type"):
                        internal_type = internal_type.of_type
                    # Get the name from the internal type if it has one
                    internal_type_name = getattr(internal_type, "name", None)
                    if internal_type_name:
                        concepts.nested_objects[field_fqn] = internal_type_name
                else:
                    # field uses a scalar type or enum type
                    concepts.objects[named_type.name].append(field_fqn)
                    concepts.fields.append(field_fqn)
                    # Enhanced metadata for advanced functionality (SKOS generation, etc.)
                    concepts.field_metadata[field_fqn] = FieldMetadata(
                        object_name=named_type.name, field_name=field_name, field_definition=field
                    )

    return concepts



"""End-to-end integration tests for Avro export with expanded instances."""

import json
from pathlib import Path
from typing import cast

import pytest
from graphql import DocumentNode

from s2dm.exporters.avro import translate_to_avro_schema
from s2dm.exporters.utils.schema_loader import load_and_process_schema


class TestAvroE2EExpandedInstances:
    @pytest.fixture
    def test_schema_path(self, spec_directory: Path) -> list[Path]:
        """Path to the test GraphQL schema."""
        return [spec_directory, Path(__file__).parent / "test_expanded_instances" / "test_schema.graphql"]

    def test_expanded_instances_default(self, test_schema_path: list[Path], tmp_path: Path) -> None:
        """Test that instance tags are NOT expanded by default (treated as regular types)."""
        query_file = tmp_path / "selection.graphql"
        query_file.write_text("query Selection { cabin { seats { isOccupied } doors { isLocked } temperature } }")

        root_type = "Cabin"
        annotated_schema, _, selection_query = load_and_process_schema(
            schema_paths=test_schema_path,
            naming_config_path=None,
            selection_query_path=query_file,
            root_type=root_type,
            expanded_instances=False,
        )
        result = translate_to_avro_schema(annotated_schema, "com.example", cast(DocumentNode, selection_query))
        result_dict = json.loads(result)

        assert result_dict["name"] == "Selection"
        assert result_dict["type"] == "record"
        assert result_dict["namespace"] == "com.example"

        cabin_field = result_dict["fields"][0]
        assert cabin_field["name"] == "cabin"
        assert cabin_field["type"][0] == "null"

        cabin_record = cabin_field["type"][1]
        assert cabin_record["type"] == "record"
        assert cabin_record["name"] == "Cabin"
        assert len(cabin_record["fields"]) == 3

        seats_field = cabin_record["fields"][0]
        assert seats_field["name"] == "seats"
        assert seats_field["type"][0] == "null"
        seats_array = seats_field["type"][1]
        assert seats_array["type"] == "array"

        seat_record = seats_array["items"][1]
        assert seat_record["type"] == "record"
        assert seat_record["name"] == "Seat"

        doors_field = cabin_record["fields"][1]
        assert doors_field["name"] == "doors"
        assert doors_field["type"][0] == "null"
        doors_array = doors_field["type"][1]
        assert doors_array["type"] == "array"

        door_record = doors_array["items"][1]
        assert door_record["type"] == "record"
        assert door_record["name"] == "Door"

        temperature_field = cabin_record["fields"][2]
        assert temperature_field["name"] == "temperature"
        assert temperature_field["type"] == ["null", "double"]

        assert "Seat_Row" not in result
        assert "Seat_Position" not in result
        assert "Door_Row" not in result
        assert "Door_Side" not in result

    def test_expanded_instances(self, test_schema_path: list[Path], tmp_path: Path) -> None:
        """Test that instance tags are expanded into nested structures when enabled."""
        query_file = tmp_path / "selection.graphql"
        query_file.write_text(
            "query Selection { cabin { seats { isOccupied height } doors { isLocked position } temperature } }"
        )

        root_type = "Cabin"
        annotated_schema, _, selection_query = load_and_process_schema(
            schema_paths=test_schema_path,
            naming_config_path=None,
            selection_query_path=query_file,
            root_type=root_type,
            expanded_instances=True,
        )
        result = translate_to_avro_schema(annotated_schema, "com.example", cast(DocumentNode, selection_query))
        result_dict = json.loads(result)

        assert result_dict["name"] == "Selection"
        assert result_dict["type"] == "record"
        assert result_dict["namespace"] == "com.example"

        cabin_field = result_dict["fields"][0]
        assert cabin_field["name"] == "cabin"
        assert isinstance(cabin_field["type"], list)
        assert cabin_field["type"][0] == "null"

        cabin_record = cabin_field["type"][1]
        assert isinstance(cabin_record, dict)
        assert cabin_record["type"] == "record"
        assert cabin_record["name"] == "Cabin"
        assert cabin_record["namespace"] == "com.example"
        assert len(cabin_record["fields"]) == 3

        temperature_field = cabin_record["fields"][0]
        assert temperature_field["name"] == "temperature"
        assert temperature_field["type"] == ["null", "double"]

        seat_field = cabin_record["fields"][1]
        assert seat_field["name"] == "Seat"
        seat_row_type = seat_field["type"]
        assert isinstance(seat_row_type, dict)
        assert seat_row_type["type"] == "record"
        assert seat_row_type["name"] == "Seat_Row"
        assert seat_row_type["namespace"] == "com.example"
        assert len(seat_row_type["fields"]) == 3

        for idx, row_name in enumerate(["ROW1", "ROW2", "ROW3"]):
            row_field = seat_row_type["fields"][idx]
            assert row_field["name"] == row_name

            if idx == 0:
                seat_position_type = row_field["type"]
                assert isinstance(seat_position_type, dict)
                assert seat_position_type["type"] == "record"
                assert seat_position_type["name"] == "Seat_Position"
                assert seat_position_type["namespace"] == "com.example"
                assert len(seat_position_type["fields"]) == 3

                for pos_idx, position_name in enumerate(["LEFT", "CENTER", "RIGHT"]):
                    position_field = seat_position_type["fields"][pos_idx]
                    assert position_field["name"] == position_name
                    assert isinstance(position_field["type"], list)
                    assert position_field["type"][0] == "null"

                    seat_type = position_field["type"][1]
                    if pos_idx == 0:
                        assert isinstance(seat_type, dict)
                        assert seat_type["type"] == "record"
                        assert seat_type["name"] == "Seat"
                        assert seat_type["namespace"] == "com.example"
                        assert len(seat_type["fields"]) == 2
                        assert seat_type["fields"][0]["name"] == "isOccupied"
                        assert seat_type["fields"][0]["type"] == ["null", "boolean"]
                        assert seat_type["fields"][1]["name"] == "height"
                        assert seat_type["fields"][1]["type"] == ["null", "int"]
                    else:
                        assert seat_type == "com.example.Seat"
            else:
                assert row_field["type"] == "com.example.Seat_Position"

        door_field = cabin_record["fields"][2]
        assert door_field["name"] == "Door"
        door_row_type = door_field["type"]
        assert isinstance(door_row_type, dict)
        assert door_row_type["type"] == "record"
        assert door_row_type["name"] == "Door_Row"
        assert door_row_type["namespace"] == "com.example"
        assert len(door_row_type["fields"]) == 2

        for idx, row_name in enumerate(["ROW1", "ROW2"]):
            door_row_field = door_row_type["fields"][idx]
            assert door_row_field["name"] == row_name

            if idx == 0:
                door_side_type = door_row_field["type"]
                assert isinstance(door_side_type, dict)
                assert door_side_type["type"] == "record"
                assert door_side_type["name"] == "Door_Side"
                assert door_side_type["namespace"] == "com.example"
                assert len(door_side_type["fields"]) == 2

                for side_idx, side_name in enumerate(["DRIVERSIDE", "PASSENGERSIDE"]):
                    side_field = door_side_type["fields"][side_idx]
                    assert side_field["name"] == side_name
                    assert isinstance(side_field["type"], list)
                    assert side_field["type"][0] == "null"

                    door_type = side_field["type"][1]
                    if side_idx == 0:
                        assert isinstance(door_type, dict)
                        assert door_type["type"] == "record"
                        assert door_type["name"] == "Door"
                        assert door_type["namespace"] == "com.example"
                        assert len(door_type["fields"]) == 2
                        assert door_type["fields"][0]["name"] == "isLocked"
                        assert door_type["fields"][0]["type"] == ["null", "boolean"]
                        assert door_type["fields"][1]["name"] == "position"
                        assert door_type["fields"][1]["type"] == ["null", "int"]
                    else:
                        assert door_type == "com.example.Door"
            else:
                assert door_row_field["type"] == "com.example.Door_Side"

    def test_expanded_instances_with_naming_config(self, test_schema_path: list[Path], tmp_path: Path) -> None:
        """Test that naming config is applied to expanded instance field names."""
        naming_config = {"field": {"object": "MACROCASE"}}
        naming_config_file = tmp_path / "naming_config.json"
        naming_config_file.write_text(json.dumps(naming_config))

        query_file = tmp_path / "selection.graphql"
        query_file.write_text(
            "query Selection { cabin { seats { isOccupied height } doors { isLocked position } temperature } }"
        )

        root_type = "Cabin"
        annotated_schema, _, selection_query = load_and_process_schema(
            schema_paths=test_schema_path,
            naming_config_path=naming_config_file,
            selection_query_path=query_file,
            root_type=root_type,
            expanded_instances=True,
        )
        result = translate_to_avro_schema(annotated_schema, "com.example", cast(DocumentNode, selection_query))
        result_dict = json.loads(result)

        assert result_dict["name"] == "Selection"
        assert result_dict["type"] == "record"
        assert result_dict["namespace"] == "com.example"

        cabin_field = result_dict["fields"][0]
        assert cabin_field["name"] == "cabin"
        assert isinstance(cabin_field["type"], list)
        assert cabin_field["type"][0] == "null"

        cabin_record = cabin_field["type"][1]
        assert isinstance(cabin_record, dict)
        assert cabin_record["type"] == "record"
        assert cabin_record["name"] == "Cabin"
        assert cabin_record["namespace"] == "com.example"
        assert len(cabin_record["fields"]) == 3

        temperature_field = cabin_record["fields"][0]
        assert temperature_field["name"] == "TEMPERATURE"
        assert temperature_field["type"] == ["null", "double"]

        seat_field = cabin_record["fields"][1]
        assert seat_field["name"] == "SEAT"
        seat_row_type = seat_field["type"]
        assert isinstance(seat_row_type, dict)
        assert seat_row_type["type"] == "record"
        assert seat_row_type["name"] == "Seat_Row"
        assert seat_row_type["namespace"] == "com.example"
        assert len(seat_row_type["fields"]) == 3

        for idx, row_name in enumerate(["ROW1", "ROW2", "ROW3"]):
            row_field = seat_row_type["fields"][idx]
            assert row_field["name"] == row_name

            if idx == 0:
                seat_position_type = row_field["type"]
                assert isinstance(seat_position_type, dict)
                assert seat_position_type["type"] == "record"
                assert seat_position_type["name"] == "Seat_Position"
                assert seat_position_type["namespace"] == "com.example"
                assert len(seat_position_type["fields"]) == 3

                for pos_idx, position_name in enumerate(["LEFT", "CENTER", "RIGHT"]):
                    position_field = seat_position_type["fields"][pos_idx]
                    assert position_field["name"] == position_name
                    assert isinstance(position_field["type"], list)
                    assert position_field["type"][0] == "null"

                    seat_type = position_field["type"][1]
                    if pos_idx == 0:
                        assert isinstance(seat_type, dict)
                        assert seat_type["type"] == "record"
                        assert seat_type["name"] == "Seat"
                        assert seat_type["namespace"] == "com.example"
                        assert len(seat_type["fields"]) == 2
                        assert seat_type["fields"][0]["name"] == "IS_OCCUPIED"
                        assert seat_type["fields"][0]["type"] == ["null", "boolean"]
                        assert seat_type["fields"][1]["name"] == "HEIGHT"
                        assert seat_type["fields"][1]["type"] == ["null", "int"]
                    else:
                        assert seat_type == "com.example.Seat"
            else:
                assert row_field["type"] == "com.example.Seat_Position"

        door_field = cabin_record["fields"][2]
        assert door_field["name"] == "DOOR"
        door_row_type = door_field["type"]
        assert isinstance(door_row_type, dict)
        assert door_row_type["type"] == "record"
        assert door_row_type["name"] == "Door_Row"
        assert door_row_type["namespace"] == "com.example"
        assert len(door_row_type["fields"]) == 2

        for idx, row_name in enumerate(["ROW1", "ROW2"]):
            door_row_field = door_row_type["fields"][idx]
            assert door_row_field["name"] == row_name

            if idx == 0:
                door_side_type = door_row_field["type"]
                assert isinstance(door_side_type, dict)
                assert door_side_type["type"] == "record"
                assert door_side_type["name"] == "Door_Side"
                assert door_side_type["namespace"] == "com.example"
                assert len(door_side_type["fields"]) == 2

                for side_idx, side_name in enumerate(["DRIVERSIDE", "PASSENGERSIDE"]):
                    side_field = door_side_type["fields"][side_idx]
                    assert side_field["name"] == side_name
                    assert isinstance(side_field["type"], list)
                    assert side_field["type"][0] == "null"

                    door_type = side_field["type"][1]
                    if side_idx == 0:
                        assert isinstance(door_type, dict)
                        assert door_type["type"] == "record"
                        assert door_type["name"] == "Door"
                        assert door_type["namespace"] == "com.example"
                        assert len(door_type["fields"]) == 2
                        assert door_type["fields"][0]["name"] == "IS_LOCKED"
                        assert door_type["fields"][0]["type"] == ["null", "boolean"]
                        assert door_type["fields"][1]["name"] == "POSITION"
                        assert door_type["fields"][1]["type"] == ["null", "int"]
                    else:
                        assert door_type == "com.example.Door"
            else:
                assert door_row_field["type"] == "com.example.Door_Side"



from pydantic import Field

from s2dm.api.models.base import QueryBasedExportRequest


class ProtobufExportRequest(QueryBasedExportRequest):
    """Request model for Protocol Buffers export endpoint."""

    flatten_naming: bool = Field(
        default=False, description="Flatten nested field names", json_schema_extra={"x-cli-flag": "--flatten-naming"}
    )
    package_name: str | None = Field(
        default=None, description="Protobuf package name", json_schema_extra={"x-cli-flag": "--package-name"}
    )



from graphql import build_schema

from s2dm.exporters.utils.naming_config import CaseFormat, NamingConventionConfig, ValidationMode
from s2dm.tools.naming_checker import check_naming_conventions


def test_type_name_violations() -> None:
    schema = build_schema("""
        type someType {
            field: String
        }

        type another_type {
            field: String
        }
    """)

    config = NamingConventionConfig.model_validate(
        {"type": {"object": CaseFormat.PASCAL_CASE}}, context={"mode": ValidationMode.CHECK}
    )
    errors = check_naming_conventions(schema, config)

    assert len(errors) == 2


def test_interface_name_violations() -> None:
    schema = build_schema("""
        interface someInterface {
            field: String
        }

        interface Another_Interface {
            field: String
        }
    """)

    config = NamingConventionConfig.model_validate(
        {"type": {"interface": CaseFormat.PASCAL_CASE}}, context={"mode": ValidationMode.CHECK}
    )
    errors = check_naming_conventions(schema, config)

    assert len(errors) == 2


def test_field_name_violations() -> None:
    schema = build_schema("""
        type SomeType {
            FieldName: String
            another_field: String
        }
    """)

    config = NamingConventionConfig.model_validate(
        {"field": {"object": CaseFormat.CAMEL_CASE}}, context={"mode": ValidationMode.CHECK}
    )
    errors = check_naming_conventions(schema, config)

    assert len(errors) == 2


def test_enum_value_violations() -> None:
    schema = build_schema("""
        enum Status {
            active
            Inactive
        }
    """)

    config = NamingConventionConfig.model_validate(
        {"enumValue": CaseFormat.MACRO_CASE}, context={"mode": ValidationMode.CHECK}
    )
    errors = check_naming_conventions(schema, config)

    assert len(errors) == 2


def test_argument_name_violations() -> None:
    schema = build_schema("""
        type Query {
            getUser(UserId: ID, user_name: String): String
        }
    """)

    config = NamingConventionConfig.model_validate(
        {"argument": {"field": CaseFormat.CAMEL_CASE}}, context={"mode": ValidationMode.CHECK}
    )
    errors = check_naming_conventions(schema, config)

    assert len(errors) == 2


def test_plural_violations() -> None:
    schema = build_schema("""
        type SomeType {
            item: [String]
            user: [String]!
        }
    """)

    config = NamingConventionConfig.model_validate({}, context={"mode": ValidationMode.CHECK})
    errors = check_naming_conventions(schema, config)

    assert len(errors) == 2


def test_no_violations() -> None:
    schema = build_schema("""
        type SomeType {
            items: [String]
            userName: String
        }

        enum Status {
            ACTIVE
            INACTIVE
        }
    """)

    config = NamingConventionConfig.model_validate(
        {
            "type": {"object": CaseFormat.PASCAL_CASE, "enum": CaseFormat.PASCAL_CASE},
            "field": {"object": CaseFormat.CAMEL_CASE},
            "enumValue": CaseFormat.MACRO_CASE,
        },
        context={"mode": ValidationMode.CHECK},
    )
    errors = check_naming_conventions(schema, config)

    assert len(errors) == 0


def test_multiple_violation_types() -> None:
    schema = build_schema("""
        type some_type {
            FieldName: String
            item: [String]
            SomeInterface: some_interface
        }

        interface some_interface {
            field: String
        }

        enum status {
            Active
        }
    """)

    config = NamingConventionConfig.model_validate(
        {
            "type": {
                "object": CaseFormat.PASCAL_CASE,
                "interface": CaseFormat.PASCAL_CASE,
                "enum": CaseFormat.PASCAL_CASE,
            },
            "field": {"object": CaseFormat.CAMEL_CASE, "interface": CaseFormat.CAMEL_CASE},
            "enumValue": CaseFormat.MACRO_CASE,
        },
        context={"mode": ValidationMode.CHECK},
    )
    errors = check_naming_conventions(schema, config)

    assert len(errors) == 7



"""Registry search helpers (e.g., SKOS search)."""

import logging
import types
from dataclasses import dataclass
from pathlib import Path
from typing import cast

from rdflib import Graph, Literal
from rdflib.plugins.sparql import prepareQuery
from rdflib.query import ResultRow

from s2dm.tools.string import normalize_whitespace

NO_LIMIT_KEYWORDS = {"inf", "infinity", "-1", "no", "none", "unlimited", "all"}


@dataclass
class SearchResult:
    """Represents a search result from SPARQL query."""

    subject: str
    predicate: str
    object_value: str
    match_type: str

    def __str__(self) -> str:
        """Return a human-readable string representation of the search result."""
        return f"{self.match_type.upper()}: {self.subject} -> {self.predicate} -> {self.object_value}"


class SKOSSearchService:
    """Service for searching SKOS RDF data using SPARQL queries."""

    def __init__(self, ttl_file_path: Path) -> None:
        """Initialize the search service with a TTL file."""
        if not ttl_file_path.exists():
            raise FileNotFoundError(f"TTL file not found: {ttl_file_path}")
        self.ttl_file_path = ttl_file_path
        self._graph: Graph | None = None

    @property
    def graph(self) -> Graph:
        """Lazy-loaded RDF graph property."""
        if self._graph is None:
            self._graph = Graph()
            try:
                self._graph.parse(self.ttl_file_path, format="turtle")
            except OSError as e:
                raise ValueError(f"Failed to read TTL file {self.ttl_file_path}: {e}") from e
            except Exception as e:
                raise ValueError(f"Failed to parse TTL file {self.ttl_file_path}: {e}") from e
            logging.info(f"Loaded {len(self._graph)} triples from {self.ttl_file_path}")
        return self._graph

    def __enter__(self) -> "SKOSSearchService":
        """Context manager entry."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        """Context manager exit with resource cleanup."""
        if self._graph is not None:
            self._graph = None

    @classmethod
    def parse_limit(cls, limit: int | str) -> int | None:
        """Parse and validate the limit parameter."""
        if isinstance(limit, str):
            if limit.lower() in NO_LIMIT_KEYWORDS:
                return None
            try:
                limit_int = int(limit)
                return limit_int if limit_int > 0 else 0
            except ValueError:
                return 0
        return limit if limit > 0 else 0

    def count_keyword_matches(self, keyword: str, ignore_case: bool = False) -> int:
        """Count total number of matches for a keyword in SKOS RDF data."""
        if not keyword.strip():
            return 0

        if ignore_case:
            subject_filter = "FILTER(CONTAINS(LCASE(STR(?subject)), LCASE(?keyword)))"
            object_filter = "FILTER(CONTAINS(LCASE(STR(?object)), LCASE(?keyword)))"
        else:
            subject_filter = "FILTER(CONTAINS(STR(?subject), ?keyword))"
            object_filter = "FILTER(CONTAINS(STR(?object), ?keyword))"

        count_query_template = f"""
        SELECT (COUNT(*) AS ?total_count)
        WHERE {{
            {{
                SELECT DISTINCT ?subject ?predicate ?object ?match_type
                WHERE {{
                    {{
                        ?subject ?predicate ?object .
                        {subject_filter}
                        BIND("subject" AS ?match_type)
                    }}
                    UNION
                    {{
                        ?subject ?predicate ?object .
                        {object_filter}
                        BIND("object" AS ?match_type)
                    }}
                }}
            }}
        }}
        """

        try:
            prepared_query = prepareQuery(count_query_template)
            results = self.graph.query(prepared_query, initBindings={"keyword": Literal(keyword)})
            for row in results:
                result_row: ResultRow = cast(ResultRow, row)
                return int(result_row["total_count"])
            return 0
        except Exception as e:
            logging.error(f"Count query execution failed: {e}")
            raise ValueError(f"Count query execution failed: {e}") from e

    def search_keyword(
        self, keyword: str, ignore_case: bool = False, limit_value: int | None = 10
    ) -> list[SearchResult]:
        """Search for a keyword in SKOS RDF data using SPARQL."""
        if not keyword.strip():
            return []
        if limit_value == 0:
            return []

        if ignore_case:
            subject_filter = "FILTER(CONTAINS(LCASE(STR(?subject)), LCASE(?keyword)))"
            object_filter = "FILTER(CONTAINS(LCASE(STR(?object)), LCASE(?keyword)))"
        else:
            subject_filter = "FILTER(CONTAINS(STR(?subject), ?keyword))"
            object_filter = "FILTER(CONTAINS(STR(?object), ?keyword))"

        query_template = f"""
        SELECT ?subject ?predicate ?object ?match_type
        WHERE {{
            {{
                ?subject ?predicate ?object .
                {subject_filter}
                BIND("subject" AS ?match_type)
            }}
            UNION
            {{
                ?subject ?predicate ?object .
                {object_filter}
                BIND("object" AS ?match_type)
            }}
        }}
        ORDER BY ?subject ?predicate
        """

        if limit_value is not None:
            query_template += f"\nLIMIT {limit_value}"

        try:
            prepared_query = prepareQuery(query_template)
        except Exception as e:
            logging.error(f"SPARQL query preparation failed: {e}")
            raise ValueError(f"Invalid SPARQL query template: {e}") from e

        try:
            results = self.graph.query(prepared_query, initBindings={"keyword": Literal(keyword)})
        except Exception as e:
            logging.error(f"SPARQL query execution failed: {e}")
            raise ValueError(f"Search query execution failed: {e}") from e

        search_results: list[SearchResult] = []
        seen_triples: set[tuple[str, str, str, str]] = set()

        for row in results:
            result_row: ResultRow = cast(ResultRow, row)
            subject = normalize_whitespace(str(result_row["subject"]))
            predicate = normalize_whitespace(str(result_row["predicate"]))
            object_value = normalize_whitespace(str(result_row["object"]))
            match_type = normalize_whitespace(str(result_row["match_type"]))

            triple_key = (subject, predicate, object_value, match_type)
            if triple_key in seen_triples:
                continue
            seen_triples.add(triple_key)
            search_results.append(
                SearchResult(
                    subject=subject,
                    predicate=predicate,
                    object_value=object_value,
                    match_type=match_type,
                )
            )

        logging.info(f"{len(search_results)} unique matches for keyword '{keyword}' with limit {limit_value}")
        return search_results
